"""
Argument parsing for yaralyze command line tool (also used by the pdfalyzer).
"""
import logging
import sys
import argparse
from argparse import _AppendAction, _StoreFalseAction, _StoreTrueAction, Action, ArgumentParser, Namespace
from importlib.metadata import version
from pathlib import Path

from rich.padding import Padding
from rich.panel import Panel
from rich.text import Text
from rich_argparse_plus import RichHelpFormatterPlus
from yaralyzer.config import YaralyzerConfig
from yaralyzer.encoding_detection.encoding_detector import CONFIDENCE_SCORE_RANGE
from yaralyzer.output.console import console
from yaralyzer.output.theme import CLI_OPTION_TYPE_STYLES, color_theme_grid
from yaralyzer.util.cli_option_validators import (DirValidator, PathValidator, OptionValidator,
     PatternsLabelValidator, YaraRegexValidator)
from yaralyzer.util.constants import *
from yaralyzer.util.exceptions import handle_argument_error
from yaralyzer.util.helpers import env_helper
from yaralyzer.util.helpers.file_helper import relative_path, timestamp_for_filename
from yaralyzer.util.helpers.shell_helper import get_inkscape_version
from yaralyzer.util.helpers.string_helper import comma_join
from yaralyzer.util.logging import highlighter, log, log_argparse_result, log_console, set_log_level
from yaralyzer.yara.yara_rule_builder import YARA_REGEX_MODIFIERS

DESCRIPTION = "Get a good hard colorful look at all the byte sequences that make up a YARA rule match."

PNG_EXPORT_WARNING = f"PNG export requires CairoSVG or Inkscape and you have neither.\n" \
                     f"Maybe try pip install {YARALYZER}[img] or {INKSCAPE_URL}"

EARLY_EXIT_ARGS = [
    '--show-colors',
    ENV_VARS_OPTION,
    '--version',
]

def epilog(config: type[YaralyzerConfig]) -> str:
    """Returns a string with some rich text tags for color to be used as the --help footer."""
    color_var = lambda s: f"[argparse.metavar]{s}[/argparse.metavar]"
    color_link = lambda s: f"[argparse.metavar]{s}[/argparse.metavar]"
    package = config.ENV_VAR_PREFIX.lower()

    msg = f"Values for most command options can be permanently set by setting via env vars or creating a " \
          f"{color_var(f'.{package}')} file. Try [argparse.args]{config.executable_name} {ENV_VARS_OPTION}" \
          f"[/argparse.args] for more info." \

    if package == YARALYZER:
        msg += f"\n[gray46]API docs: {color_link(YARALYZER_API_DOCS_URL)}[/gray46]"

    return msg


# Positional args, version, help, etc
RichHelpFormatterPlus.choose_theme('prince')  # Check options: print(RichHelpFormatterPlus.styles)
parser = ArgumentParser(formatter_class=RichHelpFormatterPlus, description=DESCRIPTION, epilog=epilog(YaralyzerConfig))

parser.add_argument('file_to_scan_path', metavar='FILE', help='file to scan', type=PathValidator())
parser.add_argument('--version', action='store_true', help='show version number and exit')

parser.add_argument('--maximize-width', action='store_true',
                    help="maximize display width to fill the terminal")

parser.add_argument(ENV_VARS_OPTION, action='store_true',
                    help=f"show env vars that can set these options permanently if placed in a .{parser.prog}r file")


# YARA rule selection
yaras = parser.add_argument_group(
    'YARA RULES',
    "Load YARA rules from preconfigured files or use one off YARA regular expression strings")

rules = yaras.add_mutually_exclusive_group(required=True)

rules.add_argument('-Y', '--yara-file',
                    help='path to a YARA rules file to check against (can be supplied more than once)',
                    action='append',
                    metavar='FILE',
                    dest='yara_rules_files',
                    type=PathValidator())

rules.add_argument('-dir', '--rule-dir',
                    help='directory with yara rules files (all files in dir are used, can be supplied more than once)',
                    action='append',
                    metavar='DIR',
                    dest='yara_rules_dirs',
                    type=DirValidator())

rules.add_argument('-re', '--regex-pattern',
                    help='build a YARA rule from PATTERN (can be supplied more than once for boolean OR)',
                    action='append',
                    metavar='PATTERN',
                    dest='regex_patterns',
                    type=YaraRegexValidator())

rules.add_argument('-hex', '--hex-pattern',
                    help='build a YARA rule from a hex string (can be supplied more than once for boolean OR)',
                    action='append',
                    metavar='HEX',
                    dest='hex_patterns',
                    type=YaraRegexValidator())

yaras.add_argument('-pl', '--patterns-label',
                    help='optional string to label your YARA patterns (makes it easier to scan results)',
                    metavar='LABEL',
                    type=PatternsLabelValidator())

yaras.add_argument('-mod', '--regex-modifier',
                    help=f"optional modifier keyword for YARA regexes ({comma_join(YARA_REGEX_MODIFIERS)})",
                    metavar='MODIFIER',
                    choices=YARA_REGEX_MODIFIERS)


# Fine tuning of output
tuning = parser.add_argument_group(
    'FINE TUNING',
    "Tune various aspects of the analyses and visualizations to your needs. As an example setting " +
        "a low --max-decode-length (or suppressing brute force binary decode attempts altogether) can " +
        "dramatically improve run times and only occasionally leads to a fatal lack of insight.")

tuning.add_argument('--surrounding-bytes',
                    help="number of bytes to display/decode before and after YARA match start positions",
                    default=YaralyzerConfig.DEFAULT_SURROUNDING_BYTES,
                    metavar='N',
                    type=int)

tuning.add_argument('--suppress-decodes-table', action='store_true',
                    help='suppress decodes table entirely (including hex/raw output)')

tuning.add_argument('--suppress-decoding-attempts', action='store_true',
                    help='suppress decode attempts for matched bytes (only hex/raw output will be shown)')

tuning.add_argument('--min-decode-length',
                    help='suppress decode attempts for quoted byte sequences shorter than N',
                    default=YaralyzerConfig.DEFAULT_MIN_DECODE_LENGTH,
                    metavar='N',
                    type=int)

tuning.add_argument('--max-decode-length',
                    help='suppress decode attempts for quoted byte sequences longer than N',
                    default=YaralyzerConfig.DEFAULT_MAX_DECODE_LENGTH,
                    metavar='N',
                    type=int)

tuning.add_argument('--suppress-chardet', action='store_true',
                    help="suppress the display of the full table of chardet's encoding likelihood scores")

tuning.add_argument('--min-chardet-bytes',
                    help="minimum number of bytes to run chardet.detect() and the decodings it suggests",
                    default=YaralyzerConfig.DEFAULT_MIN_CHARDET_BYTES,
                    metavar='N',
                    type=int)

tuning.add_argument('--min-chardet-table-confidence',
                    help="minimum chardet confidence to display the encoding name/score in the character " +
                         "decection scores table",
                    default=YaralyzerConfig.DEFAULT_MIN_CHARDET_TABLE_CONFIDENCE,
                    metavar='PCT_CONFIDENCE',
                    type=int,
                    choices=CONFIDENCE_SCORE_RANGE)

tuning.add_argument('--force-display-threshold',
                    help="encodings with chardet confidence below this number will neither be displayed nor " +
                         "decoded in the decodings table",
                    default=YaralyzerConfig.DEFAULT_FORCE_DISPLAY_THRESHOLD,
                    metavar='PCT_CONFIDENCE',
                    type=int,
                    choices=CONFIDENCE_SCORE_RANGE)

tuning.add_argument('--force-decode-threshold',
                    help="extremely high (AKA 'above this number') confidence scores from chardet.detect() " +
                         "as to the likelihood some bytes were written with a particular encoding will cause " +
                         "the yaralyzer to attempt decoding those bytes in that encoding even if it is not a " +
                         "configured encoding",
                    default=YaralyzerConfig.DEFAULT_FORCE_DECODE_THRESHOLD,
                    metavar='PCT_CONFIDENCE',
                    type=int,
                    choices=CONFIDENCE_SCORE_RANGE)

tuning.add_argument('--max-match-length',
                    help="max bytes YARA will return for a match (value is passed to YARA)",
                    default=YaralyzerConfig.DEFAULT_MAX_MATCH_LENGTH,
                    metavar='N',
                    type=int)

tuning.add_argument('--yara-stack-size',
                    help="YARA matching engine internal stack size (value is passed to YARA)",
                    default=YaralyzerConfig.DEFAULT_YARA_STACK_SIZE,
                    metavar='N',
                    type=int)


# Export options
export = parser.add_argument_group(
    'FILE EXPORT',
    "Multiselect. Choosing nothing is choosing nothing. Sends what you see on the screen to various file " +
        "formats in parallel. Writes files to the current directory if --output-dir is not provided. " +
        "Filenames are expansions of the scanned filename though you can use --file-prefix and " +
        "--file-suffix to make your rendered files more unique and beautiful to their beholder.")

export.add_argument('-html', '--export-html', action='store_const',
                    const='html',
                    help='export analysis to styled html files')

export.add_argument('-txt', '--export-txt', action='store_const',
                    const='txt',
                    help='export analysis to ANSI colored text files')

export.add_argument('-svg', '--export-svg', action='store_const',
                    const='svg',
                    help='export analysis to SVG images')

export.add_argument('-png', '--export-png', action='store_const',
                    const='png',
                    help='export analysis to PNG images (requires cairosvg or inkscape)')

export.add_argument('-json', '--export-json', action='store_const',
                    const='json',
                    help='export analysis to JSON files (experimental / possibly incomplete)')

export.add_argument('-out', '--output-dir',
                    metavar='OUTPUT_DIR',
                    help='export files to OUTPUT_DIR instead of the current directory',
                    type=DirValidator())

export.add_argument('-pfx', '--file-prefix',
                    metavar='PREFIX',
                    help='optional string to use as the prefix for exported files of any kind',
                    default='')

export.add_argument('-sfx', '--file-suffix',
                    metavar='SUFFIX',
                    help='optional string to use as the suffix for exported files of any kind',
                    default='')

export.add_argument(ECHO_COMMAND_OPTION, action='store_true',
                   help="prepend the exact command line used to the output (useful for revisiting old exports)")

export.add_argument(NO_TIMESTAMPS_OPTION, action='store_true',
                    help="do not append file creation timestamps to exported filenames")

export.add_argument(SUPPRESS_OUTPUT_OPTION, action='store_true',
                    help="no output to terminal (useful when you're exporting HTML etc.)")


# Debugging
debug = parser.add_argument_group('DEBUG', 'Debugging/interactive options.')
level = debug.add_mutually_exclusive_group()

level.add_argument('-D', '--debug', action='store_true',
                   help='show verbose debug log output')

level.add_argument('-L', '--log-level',
                   help='set the log level',
                   choices=[level for level in logging.getLevelNamesMapping().keys()])

debug.add_argument('-I', '--interact', action='store_true',
                   help='drop into interactive python REPL when parsing is complete')

debug.add_argument('--show-colors', action='store_true',
                   help='show the color theme and exit')


# Preliminaty sys.argv processing
should_exit_early = any(arg in EARLY_EXIT_ARGS for arg in sys.argv)
# TODO: replacing -rpl with to avoid breaking existing users. Delete on version bump.
sys.argv = ['-pl' if arg == '-rpl' else arg for arg in sys.argv]


def parse_arguments(config: type[YaralyzerConfig], _args: Namespace | None = None):
    """
    Parse command line args. Most arguments can also also be chosen by setting env vars,
    run with `--env-vars` option for more info on how that works.
    "Private" args injected outside of user selection will be prefixed with underscore.

    Args:
        config (type[YaralyzerConfig]): Either YarlyzeerConfig or PdfalyzerConfig.
        args (Namespace, optional): Use these `args` instead of parsing from `sys.argv`.
            Must come from an `ArgumentParser` that inherits from Yaralyzer's.

    Raises:
        InvalidArgumentError: If args are invalid.
    """
    if should_exit_early:
        if '--version' in sys.argv:
            print(f"{config.app_name} {version(config.app_name)}")
        elif ENV_VARS_OPTION in sys.argv:
            show_configurable_env_vars(config)
        elif '--show-colors' in sys.argv:
            console.print(color_theme_grid(config.COLOR_THEME, config.app_name))

        sys.exit()

    # Parse args and set a few private variables we want that are unrelated to user input
    args = _args or parser.parse_args()
    args._invoked_at_str = timestamp_for_filename()
    args._standalone_mode = _args is None
    args._any_export_selected = any(k for k, v in vars(args).items() if k.startswith('export') and v)

    if args.maximize_width:
        console.width = max(env_helper.console_width_possibilities())

    if args._any_export_selected:
        console.record = True
    elif args.output_dir:
        log.warning('--output-dir provided but no export option was chosen')

    # SVG export is a necessary intermediate step for PNG export
    if args.export_png:
        args._keep_exported_svg = bool(args.export_svg)

        if not (env_helper.is_cairosvg_installed() or get_inkscape_version()):
            handle_argument_error(PNG_EXPORT_WARNING, is_standalone_mode=args._standalone_mode)
        elif not args.export_svg:
            args.export_svg = 'svg'

    if args.debug:
        set_log_level(logging.DEBUG)
        log_argparse_result(args, 'End of yaralyzer parse_arguments()')
    elif args.log_level:
        set_log_level(args.log_level)

    return args


def show_configurable_env_vars(config: type[YaralyzerConfig]) -> None:
    """
    Show the environment variables that can be used to set command line options, either
    permanently in a `.yaralyzer` file or in other standard environment variable ways.
    """
    panel = Panel(f"{config.app_name.title()} Environment Variables", style='honeydew2')
    log_console.print(Padding(panel, (1, 0, 0, 0)), justify='center', width=int(env_helper.CONSOLE_WIDTH / 2))
    log_console.print(_configurable_env_vars_header(config.ENV_VAR_PREFIX), style='grey54')

    for group in [g for g in config._argument_parser._action_groups if 'positional' not in str(g.title)]:
        log_console.print(f"\n# {group.title}", style=RichHelpFormatterPlus.styles["argparse.groups"])

        for action in group._group_actions:
            if not config._is_configurable_by_env_var(action.dest):
                continue

            var = config.env_var_for_command_line_option(action.dest)
            _print_env_var_explanation(var, action, config)

    _print_env_var_explanation(config.log_dir_env_var, 'writing of logs to files', config)
    log_console.line()


def _configurable_env_vars_header(app_name: str) -> Padding:
    """Informational panel about the configurable env vars."""
    app_name = app_name.lower()
    txt = Text(f"These are the environment variables can be set to configure {app_name}'s command line\n"
               f"options, either by conventional environment variable setting methods or by creating\na ")
    txt.append(f"{dotfile_name(app_name)} ", style='bright_cyan bold')
    txt.append(f"file in your home or current directory and putting these vars in it.\n\n"
               f"For more on how that works see the example env file here:\n\n   ")
    txt.append(f"{example_dotenv_file_url(app_name)}", style='cornflower_blue underline bold')
    return Padding(txt, (1, 1, 0, 2))


def _print_env_var_explanation(env_var: str, action: str | Action, config: type[YaralyzerConfig]) -> None:
    """Print a line explaiing which command line option corresponds to this `env_var`."""
    env_var_style = RichHelpFormatterPlus.styles["argparse.args"].replace('italic', '')
    option = action.option_strings[-1] if isinstance(action, Action) else action

    if isinstance(action, str):
        option_type = 'Path' if env_helper.is_path_var(env_var) else 'str'
    elif isinstance(action, (_StoreFalseAction, _StoreTrueAction)):
        option_type = 'bool'
    elif isinstance(action.type, OptionValidator):
        option_type = action.type.arg_type_str()
    elif action.type is not None:
        option_type = action.type.__name__
    else:
        option_type = 'str'

    # stderr_console.print(f"env_var={env_var}, acitoncls={type(action).__name__}, action.type={action.type}")
    comment = ' (comma separated for multiple)' if isinstance(action, _AppendAction) else ''
    env_value = config.get_env_value(env_var)
    txt = Text('  ').append(f"{env_var:40}", style=env_var_style)
    txt.append(f' {option_type:8} ', style=CLI_OPTION_TYPE_STYLES.get(option_type, 'white') + ' dim italic')
    txt.append(' sets ').append(f"{option:{_max_arg_width()}} ", style='honeydew2').append(comment, style='dim')

    if (env_value := config.get_env_value(env_var)) is not None:
        env_value = [str(e) for e in env_value] if isinstance(env_value, list) else env_value
        txt += Text(f"[env: ", style='bold reverse').append(highlighter(f"{env_value}")).append(']')

    log_console.print(txt)


def _max_arg_width() -> int:
    """Maximum length of an available argument string."""
    opts = [opt for opt in parser._actions if 'option_strings' in dir(opt)]
    return max(len(o) for opt in opts for o in opt.option_strings)



import logging

from rich.text import Text
from typing import Literal


# App Strings
YARALYZE = 'yaralyze'
YARALYZER = f"{YARALYZE}r"
YARALYZER_UPPER = YARALYZER.upper()
INVOKED_BY_PYTEST = 'INVOKED_BY_PYTEST'

dotfile_name = lambda app_name: f".{app_name}".lower()

# Numbeers
KILOBYTE = 1024
MEGABYTE = KILOBYTE * KILOBYTE
MAX_FILENAME_LENGTH = 245

# Command line options
ECHO_COMMAND_OPTION = '--echo-command'
ENV_VARS_OPTION = '--env-vars'
NO_TIMESTAMPS_OPTION = '--no-timestamps'
SUPPRESS_OUTPUT_OPTION = '--suppress-output'

# Repos
GITHUB_BASE_URL = 'https://github.com/michelcrypt4d4mus'

repo_url = lambda app_name: f"{GITHUB_BASE_URL}/{app_name.lower()}"
example_dotenv_file_url = lambda app_name: f"{repo_url(app_name)}/blob/master/{dotfile_name(app_name)}.example"

PDFALYZER_REPO_URL = repo_url('pdfalyzer')
YARALYZER_REPO_URL = repo_url(YARALYZER)
YARALYZER_API_DOCS_URL = f"https://michelcrypt4d4mus.github.io/{YARALYZER}"

PDFALYZER_MSG_TXT = Text("\nIf you are analyzing a PDF you may be interested in Pdfalyzer, ", style='bright_white bold') \
                 .append("birthplace of the Yaralyzer:\n -> ").append('') \
                 .append(f'{PDFALYZER_REPO_URL}\n', style='bright_cyan underline')

# Misc URLs etc.
INKSCAPE = 'inkscape'
INKSCAPE_URL = f'https://{INKSCAPE}.org/'



"""
`BytesDecoder` class for attempting to decode bytes with various encodings.
"""
from collections import defaultdict
from copy import deepcopy
from dataclasses import dataclass, field
from operator import attrgetter

from rich.align import Align
from rich.console import Console, ConsoleOptions, NewLine, RenderResult
from rich.panel import Panel
from rich.table import Table
from rich.text import Text

from yaralyzer.bytes_match import BytesMatch  # Used to cause circular import issues
from yaralyzer.config import YaralyzerConfig
from yaralyzer.decoding.decoding_attempt import DecodingAttempt
from yaralyzer.encoding_detection.character_encodings import ENCODING, ENCODINGS_TO_ATTEMPT
from yaralyzer.encoding_detection.encoding_assessment import EncodingAssessment
from yaralyzer.encoding_detection.encoding_detector import EncodingDetector
from yaralyzer.util.helpers.collections_helper import get_dict_key_by_value
from yaralyzer.util.helpers.rich_helper import DECODING_ERRORS_MSG, NO_DECODING_ERRORS_MSG
from yaralyzer.output.decoding_attempts_table import new_decoding_attempts_table
from yaralyzer.output.decoding_table_row import DecodingTableRow
from yaralyzer.util.logging import log

# A 2-tuple that can be indexed by booleans of messages used in the table to show true vs. false
WAS_DECODABLE_YES_NO = [NO_DECODING_ERRORS_MSG, DECODING_ERRORS_MSG]
# Multiply chardet scores by 100 (again) to make sorting the table easy
SCORE_SCALER = 100.0


@dataclass
class BytesDecoder:
    """
    Handles decoding a chunk of bytes into strings using various possible encodings, ranking and displaying results.

    This class leverages the `chardet` library and custom logic to try multiple encodings, track decoding outcomes,
    and present the results in a rich, user-friendly format. It is used to analyze and display the possible
    interpretations of a byte sequence, especially in the context of YARA matches or binary analysis.

    Attributes:
        bytes_match (BytesMatch): The `BytesMatch` instance being decoded.
        label (str, optional): Label for this decoding attempt, defaults to `bytes_match.label`.
        bytes (bytes): The bytes (including surrounding context) to decode.
        decoded_strings (dict[str, str]): Maps encoding to decoded string.
        decodings (list[DecodingAttempt]): DecodingAttempt objects for each encoding tried.
        encoding_detector (EncodingDetector): Used to detect and assess possible encodings.
        was_match_decodable (dict): Tracks successful decodes per encoding.
        was_match_force_decoded (dict): Tracks forced decodes per encoding.
        was_match_undecodable (dict): Tracks failed decodes per encoding.
    """
    bytes_match: BytesMatch
    label: str = ''
    # Non-arguments
    decoded_strings: dict[str, str] = field(default_factory=dict)
    decodings: list[DecodingAttempt] = field(default_factory=list)
    encoding_detector: EncodingDetector = field(init=False)
    was_match_decodable: defaultdict[str, int] = field(default_factory=lambda: _build_encodings_metric_dict())
    was_match_force_decoded: defaultdict[str, int] = field(default_factory=lambda: _build_encodings_metric_dict())
    was_match_undecodable: defaultdict[str, int] = field(default_factory=lambda: _build_encodings_metric_dict())

    @property
    def bytes(self) -> bytes:
        return self.bytes_match.surrounding_bytes

    def __post_init__(self):
        # Note we instantiate EncodingDetector both the match and surrounding bytes
        self.encoding_detector = EncodingDetector(self.bytes)
        self.label = self.label or self.bytes_match.label

    def _build_decodings_table(self, suppress_decodes: bool = False) -> Table:
        """
        First rows are the raw / hex views of the bytes, next rows are the attempted decodings.

        Args:
            suppress_decodes (bool, optional): If `True` don't add decoding attempts to the table. Defaults to `False`.
        """
        self.table = new_decoding_attempts_table(self.bytes_match)

        # Add the encoding rows to the table if not suppressed
        if not (YaralyzerConfig.args.suppress_decoding_attempts or suppress_decodes):
            self.decodings = [DecodingAttempt(self.bytes_match, encoding) for encoding in ENCODINGS_TO_ATTEMPT]
            # Attempt decodings we don't usually attempt if chardet is insistent enough
            forced_decodes = self._undecoded_assessments(self.encoding_detector.force_decode_assessments)
            self.decodings += [DecodingAttempt(self.bytes_match, a.encoding) for a in forced_decodes]

            # If we still haven't decoded chardet's top choice, decode it
            if len(self._forced_displays()) > 0 and not self._was_decoded(self._forced_displays()[0].encoding):
                chardet_top_encoding = self._forced_displays()[0].encoding
                log.info(f"Decoding {chardet_top_encoding} because it's chardet top choice...")
                self.decodings.append(DecodingAttempt(self.bytes_match, chardet_top_encoding))

            # Build the table rows from the decoding attempts
            rows = [self._row_from_decoding_attempt(decoding) for decoding in self.decodings]

            # Add assessments with no decode attempt
            rows += [
                DecodingTableRow.from_undecoded_assessment(a, a.confidence * SCORE_SCALER)
                for a in self._forced_displays()
            ]

            self._track_decode_stats()

            for row in sorted(rows, key=attrgetter('sort_score', 'encoding_label_plain'), reverse=True):
                self.table.add_row(*row.to_row_list())

        return self.table

    # TODO: rename this to something that makes more sense, maybe assessments_over_display_threshold()?
    def _forced_displays(self) -> list[EncodingAssessment]:
        """Returns assessments over the display threshold that are not yet decoded."""
        return self._undecoded_assessments(self.encoding_detector.force_display_assessments)

    def _undecoded_assessments(self, assessments: list[EncodingAssessment]) -> list[EncodingAssessment]:
        """Filter out the already decoded assessments from a set of assessments."""
        return [a for a in assessments if not self._was_decoded(a.encoding)]

    def _was_decoded(self, encoding: str) -> bool:
        """Check whether a given encoding is in the table already."""
        return any(row.encoding == encoding for row in self.decodings)

    def _decode_attempt_subheading(self) -> Panel:
        """Generate a rich.Panel for displaying decode attempts."""
        headline = Text(f"Found ", style='decode.subheading') + self.bytes_match.__rich__()
        return Panel(headline, style='decode.subheading', expand=False)

    def _track_decode_stats(self) -> None:
        """Track stats about successful vs. forced vs. failed decode attempts."""
        for decoding in self.decodings:
            if decoding.failed_to_decode:
                self.was_match_undecodable[decoding.encoding] += 1
                continue

            self.was_match_decodable[decoding.encoding] += 1

            if decoding.was_force_decoded:
                self.was_match_force_decoded[decoding.encoding] += 1

    def _row_from_decoding_attempt(self, decoding: DecodingAttempt) -> DecodingTableRow:
        """Create a `DecodingAttemptTable` row from a `DecodingAttempt`."""
        assessment = self.encoding_detector.get_encoding_assessment(decoding.encoding)

        # If the decoding can have a start offset add an appropriate extension to the encoding label
        if decoding.start_offset_label:
            if assessment.language:
                log.warning(f"{decoding.encoding} offset {decoding.start_offset} AND language '{assessment.language}'")
            else:
                assessment = deepcopy(assessment)
                assessment.set_encoding_label(decoding.start_offset_label)

        plain_decoded_string = decoding.decoded_string.plain
        sort_score = assessment.confidence * SCORE_SCALER

        # If the decoding result is a duplicate of a previous decoding, replace the decoded text
        # with "same output as X" where X is the previous encoding that gave the same result.
        if plain_decoded_string in self.decoded_strings.values():
            encoding_with_same_output = get_dict_key_by_value(self.decoded_strings, plain_decoded_string)
            display_text = Text('same output as ', style='color(66) dim italic')
            display_text.append(encoding_with_same_output, style=ENCODING).append('...', style='white')
        else:
            self.decoded_strings[decoding.encoding_label] = plain_decoded_string
            display_text = decoding.decoded_string

        # Set failures negative, shave off a little for forced decodes
        if decoding.failed_to_decode:
            sort_score = (sort_score * -1) - 100
        elif decoding.was_force_decoded:
            sort_score -= 10

        was_forced = WAS_DECODABLE_YES_NO[int(decoding.was_force_decoded)]
        return DecodingTableRow.from_decoded_assessment(assessment, was_forced, display_text, sort_score)

    def __rich_console__(self, _console: Console, options: ConsoleOptions) -> RenderResult:
        """Rich object generator (see Rich console docs)."""
        yield NewLine(2)
        yield Align(self._decode_attempt_subheading(), 'center')

        if not YaralyzerConfig.args.suppress_chardet:
            yield NewLine()
            yield Align(self.encoding_detector, 'center')
            yield NewLine()

        # In standalone mode we always print the hex/raw bytes
        if self.bytes_match.is_decodable():
            yield self._build_decodings_table()
        elif YaralyzerConfig.args._standalone_mode:
            yield self._build_decodings_table(True)

        yield NewLine()
        yield Align(self.bytes_match.bytes_hashes_table(), 'center', style='dim')


def _build_encodings_metric_dict():
    """One key for each key in `ENCODINGS_TO_ATTEMPT`, values are all 0."""
    metrics_dict = defaultdict(lambda: 0)

    for encoding in ENCODINGS_TO_ATTEMPT.keys():
        metrics_dict[encoding] = 0

    return metrics_dict



"""
Builds bare bones YARA rules to match strings and regex patterns.

Example rule string:

```
rule Just_A_Piano_Man {
    meta:
        author           = "Tim"
    strings:
        $hilton_producer = /Scott.*Storch/
    condition:
        $hilton_producer
}
```
"""
import re
from typing import Literal, Optional

import yara

from yaralyzer.util.constants import YARALYZE
from yaralyzer.util.logging import log

PatternType = Literal['hex', 'regex']
YaraModifierType = Literal['ascii', 'fullword', 'nocase', 'wide']

HEX = 'hex'
PATTERN = 'pattern'
REGEX = 'regex'
RULE = 'rule'
UNDERSCORE = '_'
YARA_REGEX_MODIFIERS = ['nocase', 'ascii', 'wide', 'fullword']
PATTERN_TYPES: list[PatternType] = ['hex', 'regex']

SAFE_LABEL_REPLACEMENTS = {
    '/': 'frontslash',
    '\\': 'backslash',
    "'": 'singlequote',
    '"': 'doublequote',
    '`': 'backtick',
    '-': UNDERSCORE,
    ' ': UNDERSCORE,
}

RULE_TEMPLATE = """
rule {rule_name} {{
    meta:
        author = "The Yaralyzer"
    strings:
        ${pattern_label} = {pattern}
    condition:
        ${pattern_label}
}}
"""

BYTES_RULE_TEMPLATE = """
rule {rule_name} {{
    meta:
        author = "The Yaralyzer"
    strings:
        ${pattern_label} = {{ {bytes_pattern} }}
    condition:
        ${pattern_label}
}}
"""


def yara_rule_string(
    pattern: str,
    pattern_type: PatternType = REGEX,
    rule_name: str = YARALYZE,
    pattern_label: Optional[str] = PATTERN,
    modifier: Optional[YaraModifierType] = None
) -> str:
    """
    Build a YARA rule string for a given `pattern`.

    Args:
        pattern (str): The string or regex pattern to match.
        pattern_type (str): Either `"regex"` or `"hex"`. Default is `"regex"`.
        rule_name (str): The name of the YARA rule. Default is `"YARALYZE"`.
        pattern_label (Optional[str]): The label for the pattern in the YARA rule. Default is `"pattern"`.
        modifier (Optional[str]): Optional regex modifier (e.g. 'nocase', 'ascii', 'wide', 'fullword').
            Only valid if `pattern_type` is `"regex"`.

    Returns:
        str: The constructed YARA rule as a string.
    """
    if not (modifier is None or modifier in YARA_REGEX_MODIFIERS):
        raise TypeError(f"Modifier '{modifier}' is not one of {YARA_REGEX_MODIFIERS}")

    if pattern_type == REGEX:
        pattern = f"/{pattern}/"
    elif pattern_type == HEX:
        pattern = f"{{{pattern}}}"
    else:
        raise ValueError(f"pattern_type must be either '{REGEX}' or '{HEX}'")

    if modifier:
        pattern += f" {modifier}"

    rule = RULE_TEMPLATE.format(
        rule_name=rule_name,
        pattern_label=pattern_label,
        pattern=pattern,
        modifier='' if modifier is None else f" {modifier}"
    )

    log.debug(f"Built YARA rule: \n{rule}")
    return rule


def build_yara_rule(
    pattern: str,
    pattern_type: PatternType = REGEX,
    rule_name: str = YARALYZE,
    pattern_label: Optional[str] = PATTERN,
    modifier: Optional[YaraModifierType] = None
) -> yara.Rule:
    """
    Build a compiled `yara.Rule` object.

    Args:
        pattern (str): The string or regex pattern to match.
        pattern_type (str): Either `"regex"` or `"hex"`. Default is `"regex"`.
        rule_name (str): The name of the YARA rule. Default is `"YARALYZE"`.
        pattern_label (Optional[str]): The label for the pattern in the YARA rule. Default is `"pattern"`.
        modifier (Optional[str]): Optional regex modifier (e.g. 'nocase', 'ascii', 'wide', 'fullword').
            Only valid if `pattern_type` is `"regex"`.

    Returns:
        yara.Rule: Compiled YARA rule object.
    """
    rule_string = yara_rule_string(pattern, pattern_type, rule_name, pattern_label, modifier)
    return yara.compile(source=rule_string)


def safe_label(_label: str) -> str:
    """
    YARA rule and pattern names can only contain alphanumeric chars.

    Args:
        _label (str): The label to sanitize.

    Returns:
        str: A sanitized label safe for use in YARA rules.
    """
    label = _label

    for char, replacement in SAFE_LABEL_REPLACEMENTS.items():
        if replacement != UNDERSCORE:
            label = label.replace(char, f"__{replacement.upper()}__")
        else:
            label = label.replace(char, replacement)

    if re.match('^\\d', label):
        label = '_' + label

    if not re.match('\\w+', label):
        msg = f"'{label}' is invalid: YARA labels must be alphanumeric/underscore and cannot start with a number"
        raise ValueError(msg)

    log.debug(f"Built safe label {label} from {_label}")
    return label



import code
import yara as python_yara
from pathlib import Path

from yaralyzer.config import YaralyzerConfig
from yaralyzer.output.console import console
from yaralyzer.output.file_export import export_json, invoke_rich_export
from yaralyzer.util.argument_parser import parser, parse_arguments
from yaralyzer.util.constants import PDFALYZER_MSG_TXT, YARALYZER
from yaralyzer.util.exceptions import print_fatal_error_and_exit
from yaralyzer.util.helpers.env_helper import load_dotenv_file
from yaralyzer.util.logging import invocation_txt, log, log_console, log_current_config
from yaralyzer.yara.error import yara_error_msg
from yaralyzer.yara.yara_rule_builder import HEX, REGEX
from yaralyzer.yaralyzer import Yaralyzer

YaralyzerConfig.set_parsers(parser, parse_arguments)


def yaralyze():
    """
    Entry point for Yaralyzer when invoked as a script. Args are parsed from the command line
    and environment variables. See `yaralyze --help` for details.
    """
    args = YaralyzerConfig.parse_args()

    if args._standalone_mode:
        log_current_config(YaralyzerConfig)

    if args.yara_rules_files:
        yaralyzer = Yaralyzer.for_rules_files(args.yara_rules_files, args.file_to_scan_path)
    elif args.yara_rules_dirs:
        yaralyzer = Yaralyzer.for_rules_dirs(args.yara_rules_dirs, args.file_to_scan_path)
    elif args.regex_patterns or args.hex_patterns:
        yaralyzer = Yaralyzer.for_patterns(
            args.regex_patterns or args.hex_patterns,
            HEX if args.hex_patterns else REGEX,
            args.file_to_scan_path,
            pattern_label=args.patterns_label,
            regex_modifier=args.regex_modifier
        )
    else:
        raise RuntimeError("No pattern or YARA file to scan against.")

    args._export_basepath = yaralyzer.export_basepath()

    if args.echo_command:
        console.print(invocation_txt())

    try:
        yaralyzer.yaralyze()
    except python_yara.Error as e:
        print_fatal_error_and_exit(yara_error_msg(e))

    if args.export_txt:
        invoke_rich_export(console.save_text, args)
    if args.export_html:
        invoke_rich_export(console.save_html, args)
    if args.export_svg:
        invoke_rich_export(console.save_svg, args)
    if args.export_json:
        export_json(yaralyzer, args)

    if str(args.file_to_scan_path).lower().endswith('pdf'):
        log_console.print(PDFALYZER_MSG_TXT)

    # Drop into interactive shell if requested
    if args.interact:
        code.interact(local=locals())



"""
Rich text decorator for YARA match dicts.

A YARA match is returned as a `dict` with this structure:

Example:
    ```
    {
        'tags': ['foo', 'bar'],
        'matches': True,
        'namespace': 'default',
        'rule': 'my_rule',
        'meta': {},
        'strings': [
            StringMatch1,
            StringMatch2
        ]
    }
    ```
"""
import re
from numbers import Number
from typing import Any, Dict

from rich.console import Console, ConsoleOptions, RenderResult
from rich.padding import Padding
from rich.panel import Panel
from rich.text import Text
from yara import StringMatch

from yaralyzer.output.console import console_width
from yaralyzer.output.theme import theme_colors_with_prefix
from yaralyzer.util.helpers.bytes_helper import clean_byte_string
from yaralyzer.util.helpers.string_helper import INDENT_SPACES
from yaralyzer.util.logging import log

MATCH_PADDING = (0, 0, 0, 1)

DATE_REGEX = re.compile('\\d{4}-\\d{2}-\\d{2}')
DIGITS_REGEX = re.compile("^\\d+$")
HEX_REGEX = re.compile('^[0-9A-Fa-f]+$')
MATCHER_VAR_REGEX = re.compile('\\$[a-z_]+')
URL_REGEX = re.compile('^https?:')

YARA_STRING_STYLES: Dict[re.Pattern, str] = {
    URL_REGEX: 'yara.url',
    DIGITS_REGEX: 'yara.number',
    HEX_REGEX: 'yara.hex',
    DATE_REGEX: 'yara.date',
    MATCHER_VAR_REGEX: 'yara.match_var'
}

RAW_YARA_THEME_COLORS = [color[len('yara') + 1:] for color in theme_colors_with_prefix('yara')]
RAW_YARA_THEME_TXT = Text('\nColor Code: ') + Text(' ').join(RAW_YARA_THEME_COLORS)
RAW_YARA_THEME_TXT.justify = 'center'


class YaraMatch:
    """Rich text decorator for YARA match dicts."""

    def __init__(self, match: dict, matched_against_bytes_label: Text) -> None:
        """
        Args:
            match (dict): The YARA match dict.
            matched_against_bytes_label (Text): Label indicating what bytes were matched against.
        """
        self.match = match
        self.rule_name = match['rule']
        self.label = matched_against_bytes_label.copy().append(f" matched rule: '", style='matched_rule')
        self.label.append(self.rule_name, style='on bright_red bold').append("'!", style='siren')

    def __rich_console__(self, _console: Console, options: ConsoleOptions) -> RenderResult:
        """Renders a rich `Panel` showing the color highlighted raw YARA match info."""
        yield Text("\n")
        yield Padding(Panel(self.label, expand=False, style=f"on color(251) reverse"), MATCH_PADDING)
        yield RAW_YARA_THEME_TXT
        yield Padding(Panel(_rich_yara_match(self.match)), MATCH_PADDING)


def _rich_yara_match(element: Any, depth: int = 0) -> Text:
    """
    Painful/hacky way of recursively coloring a YARA match dict.

    Args:
        element (Any): The element to render (can be `dict`, `list`, `str`, `bytes`, `int`, `bool`).
        depth (int): Current recursion depth (used for indentation).

    Returns:
        Text: The rich `Text` representation of the element.
    """
    indent = Text((depth + 1) * INDENT_SPACES)
    end_indent = Text(depth * INDENT_SPACES)

    if isinstance(element, str):
        txt = _yara_string(element)
    elif isinstance(element, bytes):
        txt = Text(clean_byte_string(element), style='bytes')
    elif isinstance(element, Number):
        txt = Text(str(element), style='bright_cyan')
    elif isinstance(element, bool):
        txt = Text(str(element), style='red' if not element else 'green')
    elif isinstance(element, (list, tuple)):
        if len(element) == 0:
            txt = Text('[]', style='white')
        else:
            if isinstance(element[0], StringMatch):
                # In yara-python 4.3.0 the StringMatch type was introduced so we just make it look like
                # the old list of tuples format (see: https://github.com/VirusTotal/yara-python/releases/tag/v4.3.0)
                match_tuples = [
                    (match.identifier, match_instance.offset, match_instance.matched_data)
                    for match in element
                    for match_instance in match.instances
                ]

                return _rich_yara_match(match_tuples, depth)

            total_length = sum([len(str(e)) for e in element]) + ((len(element) - 1) * 2) + len(indent) + 2
            elements_txt = [_rich_yara_match(e, depth + 1) for e in element]
            list_txt = Text('[', style='white')

            if total_length > console_width() or len(element) > 3:
                join_txt = Text(f"\n{indent}")
                list_txt.append(join_txt).append(Text(f",{join_txt}").join(elements_txt))
                list_txt += Text(f'\n{end_indent}]', style='white')
            else:
                list_txt += Text(', ').join(elements_txt) + Text(']')

            return list_txt
    elif isinstance(element, dict):
        element = {k: v for k, v in element.items() if k not in ['matches', 'rule']}

        if len(element) == 0:
            return Text('{}')

        txt = Text('{\n', style='white')

        for i, k in enumerate(element.keys()):
            v = element[k]
            txt += indent + Text(f"{k}: ", style='yara.key') + _rich_yara_match(v, depth + 1)

            if (i + 1) < len(element.keys()):
                txt.append(",\n")
            else:
                txt.append("\n")

        txt += end_indent + Text('}', style='white')
    else:
        log.warning(f"Unknown yara return of type {type(element)}: {element}")
        txt = indent + Text(str(element))

    return txt


def _yara_string(_string: str) -> Text:
    """Apply special styles to certain types of yara strings (e.g. URLs, numbers, hex, dates, matcher vars)."""
    for regex in YARA_STRING_STYLES.keys():
        if regex.match(_string):
            return Text(_string, YARA_STRING_STYLES[regex])

    return Text(_string, style='yara.string')



"""
Help with lists and dicts.
"""


def flatten(a: list) -> list:
    """
    Flatten a list that may contain other lists which may also contain lists etc.
    From https://www.geeksforgeeks.org/python/python-flatten-list-to-individual-elements/.
    """
    return_value = []

    for x in a:
        if isinstance(x, list):
            return_value.extend(flatten(x))  # Recursively flatten nested lists
        else:
            return_value.append(x)  # Append individual elements

    return return_value


def get_dict_key_by_value(_dict: dict, value):
    """Inverse of the usual dict operation."""
    return list(_dict.keys())[list(_dict.values()).index(value)]


def listify(obj: object) -> list:
    """Make sure `obj` is a list."""
    return obj if isinstance(obj, list) else [obj]



"""
Helper methods to work with bytes.
"""
import re
from io import StringIO
from sys import byteorder

from rich.console import Console
from rich.markup import escape
from rich.padding import Padding
from rich.text import Text

from yaralyzer.bytes_match import BytesMatch
from yaralyzer.config import YaralyzerConfig
from yaralyzer.encoding_detection.character_encodings import NEWLINE_BYTE, encoding_width
from yaralyzer.output.console import console, console_width
from yaralyzer.output.theme import BYTES, BYTES_BRIGHTER, BYTES_BRIGHTEST, BYTES_HIGHLIGHT, GREY
from yaralyzer.util.helpers.rich_helper import newline_join
from yaralyzer.util.logging import log

HEX_CHARS_PER_GROUP = 8
SUBTABLE_MAX_WIDTH = console_width() - 35 - 5  # 35 for first 3 cols, 5 for in between hex and ascii
HEX_UNIT_LENGTH = (HEX_CHARS_PER_GROUP * 3) + HEX_CHARS_PER_GROUP + 4  # 4 for padding between groups
HEX_GROUPS_PER_LINE = divmod(SUBTABLE_MAX_WIDTH, HEX_UNIT_LENGTH)[0]
HEX_CHARS_PER_LINE = HEX_CHARS_PER_GROUP * HEX_GROUPS_PER_LINE


def get_bytes_before_and_after_match(_bytes: bytes, match: re.Match, num_before=None, num_after=None) -> bytes:
    """
    Get bytes before and after a regex match within a byte sequence.

    Args:
        _bytes (bytes): The full byte sequence.
        match (re.Match): The regex `Match` object.
        num_before (int, optional): Number of bytes before the match to include. Defaults to configured value.
        num_after (int, optional): Number of bytes after the match to include. Defaults to either configured value
            or the `num_before` arg value.

    Returns:
        bytes: The surrounding bytes including the match.
    """
    return get_bytes_surrounding_range(_bytes, match.start(), match.end(), num_before, num_after)


def get_bytes_surrounding_range(_bytes: bytes, start_idx: int, end_idx: int, num_before=None, num_after=None) -> bytes:
    """
    Get bytes surrounding a specified range in a byte sequence.

    Args:
        _bytes (bytes): The full byte sequence.
        start_idx (int): Start index of the range.
        end_idx (int): End index of the range.
        num_before (int, optional): Number of bytes before the range. Defaults to configured value.
        num_after (int, optional): Number of bytes after the range. Defaults to configured value.

    Returns:
        bytes: The surrounding bytes including the range.
    """
    num_after = num_after or num_before or YaralyzerConfig.args.surrounding_bytes
    num_before = num_before or YaralyzerConfig.args.surrounding_bytes
    start_idx = max(start_idx - num_before, 0)
    end_idx = min(end_idx + num_after, len(_bytes))
    return _bytes[start_idx:end_idx]


def clean_byte_string(bytes_array: bytes) -> str:
    r"""
    Return a clean string representation of bytes, without Python's b'' or b"" wrappers.
    e.g. '\x80\nx44' instead of "b'\x80\nx44'".

    Args:
        bytes_array (bytes): The bytes to convert.

    Returns:
        str: Clean string representation of the bytes.
    """
    byte_printer = Console(file=StringIO())
    byte_printer.out(bytes_array, end='')
    bytestr = byte_printer.file.getvalue()

    if bytestr.startswith("b'"):
        bytestr = bytestr.removeprefix("b'").removesuffix("'")
    elif bytestr.startswith('b"'):
        bytestr = bytestr.removeprefix('b"').removesuffix('"')
    else:
        raise RuntimeError(f"Unexpected byte string {bytestr}")

    return bytestr


def rich_text_view_of_raw_bytes(_bytes: bytes, bytes_match: BytesMatch) -> Text:
    """
    Return a rich `Text` object of raw bytes, highlighting the matched bytes.

    Args:
        _bytes (bytes): The full byte sequence.
        bytes_match (BytesMatch): The BytesMatch object indicating which bytes to highlight.

    Returns:
        Text: Rich Text object with highlighted match.
    """
    surrounding_bytes_str = clean_byte_string(_bytes)
    highlighted_bytes_str = clean_byte_string(bytes_match.bytes)
    highlighted_bytes_str_length = len(highlighted_bytes_str)
    highlight_idx = _find_str_rep_of_bytes(surrounding_bytes_str, highlighted_bytes_str, bytes_match)

    txt = Text(surrounding_bytes_str[:highlight_idx], style=GREY)
    matched_bytes_str = surrounding_bytes_str[highlight_idx:highlight_idx + highlighted_bytes_str_length]
    txt.append(matched_bytes_str, style=bytes_match.highlight_style)
    txt.append(surrounding_bytes_str[highlight_idx + highlighted_bytes_str_length:], style=GREY)
    return txt


def hex_view_of_raw_bytes(_bytes: bytes, bytes_match: BytesMatch) -> Text:
    """
    Return a hexadecimal view of raw bytes, highlighting the matched bytes.

    Args:
        _bytes (bytes): The full byte sequence.
        bytes_match (BytesMatch): The BytesMatch object indicating which bytes to highlight.

    Returns:
        Text: Rich Text object with highlighted match in hex view.
    """
    hex_str = hex_text(_bytes)
    highlight_start_idx = bytes_match.highlight_start_idx * 3
    highlight_end_idx = bytes_match.highlight_end_idx * 3
    hex_str.stylize(bytes_match.highlight_style, highlight_start_idx, highlight_end_idx)
    lines = hex_str.wrap(console, HEX_CHARS_PER_LINE * 3)
    return newline_join([Text('  ').join(line.wrap(console, HEX_CHARS_PER_GROUP * 3)) for line in lines])


def ascii_view_of_raw_bytes(_bytes: bytes, bytes_match: BytesMatch) -> Text:
    """
    Return an ASCII view of raw bytes, highlighting the matched bytes.

    Args:
        _bytes (bytes): The full byte sequence.
        bytes_match (BytesMatch): The BytesMatch object indicating which bytes to highlight.

    Returns:
        Text: Rich Text object with highlighted match in ASCII view.
    """
    txt = Text('', style=BYTES)

    for i, b in enumerate(_bytes):
        if i < bytes_match.highlight_start_idx or i > bytes_match.highlight_end_idx:
            style1 = 'color(246)'
            style2 = 'color(234)'
        else:
            style1 = None
            style2 = None

        _byte = b.to_bytes(1, byteorder)

        if b < 32:
            txt.append('*', style=style2 or BYTES_BRIGHTER)
        elif b < 127:
            txt.append(_byte.decode('UTF-8'), style1 or BYTES_BRIGHTEST)
        elif b <= 160:
            txt.append('*', style=style2 or BYTES_HIGHLIGHT)
        else:
            txt.append('*', style=style2 or BYTES)

    segments = [txt[i:i + HEX_CHARS_PER_GROUP] for i in range(0, len(txt), HEX_CHARS_PER_GROUP)]

    lines = [
        Text('  ').join(segments[i:min(len(segments), i + HEX_GROUPS_PER_LINE)])
        for i in range(0, len(segments), HEX_GROUPS_PER_LINE)
    ]

    return newline_join(lines)


def hex_text(_bytes: bytes) -> Text:
    """
    Return a rich Text object of the hex string for the given bytes.

    Args:
        _bytes (bytes): The bytes to convert.

    Returns:
        Text: Rich Text object of the hex string.
    """
    return Text(hex_string(_bytes), style=GREY)


def hex_string(_bytes: bytes) -> str:
    """
    Return a hex string representation of the given bytes.

    Args:
        _bytes (bytes): The bytes to convert.

    Returns:
        str: Hex string representation of the bytes.
    """
    return ' '.join([hex(b).removeprefix('0x').rjust(2, '0') for i, b in enumerate(_bytes)])


def print_bytes(bytes_array: bytes, style: str | None = None, indent: int = 0) -> None:
    """
    Print a string representation of some bytes to the console.

    Args:
        bytes_array (bytes): The bytes to print.
        style (str, optional): Style to use for printing. Defaults to 'bytes'.
    """
    for line in bytes_array.split(NEWLINE_BYTE):
        padded_bytes = Padding(escape(clean_byte_string(line)), (0, 0, 0, indent))
        console.print(padded_bytes, style=style or 'bytes')


def truncate_for_encoding(_bytes: bytes, encoding: str) -> bytes:
    """
    Truncate bytes to a multiple of the character width for the given encoding.
    For example, for utf-16 this means truncating to a multiple of 2, for utf-32 to a multiple of 4.

    Args:
        _bytes (bytes): The bytes to truncate.
        encoding (str): The encoding to consider.

    Returns:
        bytes: Truncated bytes.
    """
    char_width = encoding_width(encoding)
    num_bytes = len(_bytes)
    num_extra_bytes = num_bytes % char_width

    if char_width <= 1 or num_bytes <= char_width or num_extra_bytes == 0:
        return _bytes
    else:
        return _bytes[:-num_extra_bytes]


def _find_str_rep_of_bytes(surrounding_bytes_str: str, highlighted_bytes_str: str, highlighted_bytes: BytesMatch):
    r"""
    Find the position of the highlighted bytes string within the surrounding bytes string.

    Both arguments are string representations of binary data. This is needed because the string
    representation of bytes can be longer than the actual bytes (e.g., '\\xcc' is 4 chars for 1 byte).

    Args:
        surrounding_bytes_str (str): String representation of the full byte sequence.
        highlighted_bytes_str (str): String representation of the matched bytes.
        highlighted_bytes (BytesMatch): The BytesMatch object for context.

    Returns:
        int: The index in the surrounding string where the highlighted bytes start, or -1 if not found.
    """
    # Start a few chars in to avoid errors: sometimes we're searching for 1 or 2 bytes and there's a false positive
    # in the extra bytes. This isn't perfect - it's starting us at the first index into the *bytes* that's safe to
    # check but this is almost certainly too soon given the large % of bytes that take 4 chars to print ('\x02' etc)
    highlight_idx = surrounding_bytes_str.find(highlighted_bytes_str, highlighted_bytes.highlight_start_idx)

    # TODO: Somehow \' and ' don't always come out the same :(
    if highlight_idx == -1:
        log.info(f"Failed to find highlighted_bytes in first pass so deleting single quotes and retrying. " +
                  "Highlighting may be off by a few chars,")

        surrounding_bytes_str = surrounding_bytes_str.replace("\\'", "'")
        highlight_idx = surrounding_bytes_str.find(highlighted_bytes_str)

        if highlight_idx == -1:
            log.warning(f"Failed to find\n{highlighted_bytes_str}\nin surrounding bytes:\n{surrounding_bytes_str}")
            log.warning("Highlighting will not work on this decoded string.")

    return highlight_idx



"""
Class to manage attempting to decode a chunk of bytes into strings with a given encoding.
"""
from dataclasses import dataclass, field
from sys import byteorder
from typing import Optional

from rich.markup import escape
from rich.text import Text

from yaralyzer.bytes_match import BytesMatch  # Formerly caused circular import issues
from yaralyzer.encoding_detection.character_encodings import (ENCODINGS_TO_ATTEMPT, SINGLE_BYTE_ENCODINGS,
     UTF_8, encoding_width, is_wide_utf)
from yaralyzer.output.theme import ALERT_STYLE, BYTES_BRIGHTER, BYTES_BRIGHTEST, BYTES_NO_DIM, GREY_ADDRESS
from yaralyzer.util.helpers.bytes_helper import clean_byte_string, truncate_for_encoding
from yaralyzer.util.helpers.rich_helper import prefix_with_style, unprintable_byte_to_text
from yaralyzer.util.logging import log


@dataclass
class DecodingAttempt:
    """
    Manages the process of attempting to decode a chunk of bytes into a string using a specified encoding.

    This class tries to decode the bytes using the provided encoding, handling both standard and custom decoding
    strategies (including multi-byte encodings and forced decoding attempts). It tracks the outcome, highlights
    the decoded output, and provides information about the decoding process.

    Attributes:
        bytes (bytes): The bytes (including context) to decode.
        bytes_match (BytesMatch): The `BytesMatch` object containing match and context info.
        encoding (str): The encoding to attempt.
        encoding_label (str): Label for the encoding (may include offset info).
        start_offset (int): Byte offset used for decoding (for multi-byte encodings).
        start_offset_label (Optional[str]): String label for the offset, if used.
        was_force_decoded (bool): True if a forced decode was attempted.
        failed_to_decode (bool): True if decoding failed.
        decoded_string (Text): The decoded string as a Rich `Text` object (with highlighting).
    """
    bytes_match: BytesMatch
    encoding: str
    # Non-args
    encoding_label: str = field(init=False)
    failed_to_decode: bool = False
    start_offset: int = 0
    start_offset_label: str | None = None
    was_force_decoded: bool = False

    @property
    def bytes(self) -> bytes:
        return self.bytes_match.surrounding_bytes

    def __post_init__(self):
        # Inferred / derived values
        self.encoding_label = self.encoding
        self.start_offset = 0  # Offset in bytes to start decoding from
        self.start_offset_label = None  # String to indicate what offset we were able to decode
        self.was_force_decoded = False
        self.failed_to_decode = False
        self.decoded_string = self._decode_bytes()

    def _decode_bytes(self) -> Text:
        """
        Tries builtin decode, hands off to other methods for harsher treatment (byte shifting for
        UTF-16/32 and custom decode for the rest) if that fails. Has side effect of setting
        `self.decoded_string` value.
        """
        try:
            decoded_string = self._to_rich_text(escape(self.bytes.decode(self.encoding)))
            log.info(f"{self.encoding} auto-decoded {self.bytes_match}")
            return decoded_string
        except UnicodeDecodeError:
            log.info(f"{self.encoding} failed on 1st pass decoding {self.bytes_match} capture; custom decoding...")
        except LookupError as e:
            log.warning(f"Unknown encoding: {self.encoding}. {e}")
            return self._failed_to_decode_msg_txt(e)

        self.was_force_decoded = True

        if is_wide_utf(self.encoding):
            return self._decode_utf_multibyte()
        else:
            return self._custom_utf_decode()

    def _custom_utf_decode(self) -> Text:
        """
        Returns a `Text` obj representing an attempt to force a UTF-8 encoding onto an array of bytes.
        """
        log.info(f"Custom decoding {self.bytes_match} with {self.encoding}...")
        unprintable_char_map = ENCODINGS_TO_ATTEMPT.get(self.encoding)
        output = Text('', style='bytes.decoded')

        # We use this to skip over bytes consumed by multi-byte UTF-n chars
        skip_next = 0

        for i, b in enumerate(self.bytes):
            if skip_next > 0:
                skip_next -= 1
                continue

            _byte = b.to_bytes(1, byteorder)

            # Color the before and after bytes grey
            if i < self.bytes_match.highlight_start_idx or i > self.bytes_match.highlight_end_idx:
                style = GREY_ADDRESS
            else:
                style = self.bytes_match.highlight_style

            if style not in [GREY_ADDRESS, ALERT_STYLE]:
                if b <= 126:
                    style = BYTES_NO_DIM
                elif b <= 192:
                    style = BYTES_BRIGHTER
                else:
                    style = BYTES_BRIGHTEST

            try:
                if unprintable_char_map is not None and b in unprintable_char_map:
                    output.append(unprintable_byte_to_text(unprintable_char_map[b], style=style))
                elif b < 127:
                    output.append(_byte.decode(self.encoding), style=style)
                elif self.encoding != UTF_8:
                    output.append(_byte.decode(self.encoding), style=style)
                # At this point we know it's UTF-8, so it must be a continuation byte
                elif b <= 192:
                    # In UTF-8 bytes from 128 to 192 is a continuation byte
                    output.append(unprintable_byte_to_text(f"CHAR{b}", style=style))
                else:
                    if b <= 223:
                        char_width = 2
                    elif b <= 239:
                        char_width = 3
                    else:
                        char_width = 4

                    wide_char = self.bytes[i:i + char_width].decode(self.encoding)
                    output.append(wide_char, style=style)
                    skip_next = char_width - 1  # Won't be set if there's a decoding exception
                    log.info(f"Skipping next {skip_next} bytes because UTF-8 multibyte char '{wide_char}' used them")
            except UnicodeDecodeError:
                output.append(clean_byte_string(_byte), style=style)

        return output

    def _decode_utf_multibyte(self) -> Text:
        """
        UTF-16/32 are fixed width and multibyte and therefore depend on the position of the starting byte
        so we try several offsets until we find one that at least kind of works.

        Returns:
            Text: Rich `Text` object representing the decoded string with highlighting.
        """
        char_width = encoding_width(self.encoding)
        last_exception = None
        decoded_str = None
        bytes_offset = 0

        # Iterate through the possibly byte offsets until we find a valid decoded string (or don't)
        while bytes_offset < char_width:
            try:
                decoded_str = truncate_for_encoding(self.bytes[bytes_offset:], self.encoding).decode(self.encoding)
            except UnicodeDecodeError as e:
                log.info(f"Exception decoding w/offset {bytes_offset} in {self.encoding}: {e}")
                last_exception = e

            # Append the current bytes_offset to the encoding label if we found a valid decoded string
            if decoded_str is not None:
                log.debug(f"Successfully decoded '{self.encoding}' w/offset {bytes_offset}")
                self.start_offset = bytes_offset
                self.start_offset_label = f"offset {self.start_offset} byte" + ('s' if self.start_offset > 1 else '')
                self.encoding_label = f"{self.encoding} ({self.start_offset_label})"
                break

            bytes_offset += 1

        if decoded_str is not None:
            return self._to_rich_text(decoded_str, bytes_offset)
        else:
            return self._failed_to_decode_msg_txt(last_exception)

    def _to_rich_text(self, _string: str, bytes_offset: int = 0) -> Text:
        """
        Convert a decoded string to highlighted `Text` representation.

        Args:
            _string (str): The decoded string to convert.
            bytes_offset (int): The byte offset used during decoding (for multi-byte encodings).
        Returns:
            Text: The rich `Text` representation of the decoded string with appropriate highlighting.
        """
        # Adjust where we start the highlighting given the multibyte nature of the encodings
        log.debug(f"Stepping through {self.encoding} encoded string...")
        txt = Text('', style=self.bytes_match.style_at_position(0))
        current_byte_idx = 0

        # Prevent unprintable chars other than newline. Some of them disfigure the terminal output permanently
        if self.encoding in SINGLE_BYTE_ENCODINGS:
            is_single_byte_encoding = True
            unprintable_chars = ENCODINGS_TO_ATTEMPT[self.encoding]
        else:
            is_single_byte_encoding = False
            unprintable_chars = {}

        for _i, c in enumerate(_string):
            char_bytes = bytes(c, self.encoding)
            char_width = len(char_bytes)
            style = self.bytes_match.style_at_position(current_byte_idx + bytes_offset)

            # 10 is newline in single byte encodings
            if c.isprintable() or (ord(c) == 10 and is_single_byte_encoding):
                txt.append(c, style)
            elif ord(c) == 9 and is_single_byte_encoding:
                txt.append(unprintable_byte_to_text('\\t', style=style))
            elif ord(c) in unprintable_chars:
                txt.append(unprintable_byte_to_text(unprintable_chars[ord(c)], style=style))
            else:
                txt.append(unprintable_byte_to_text(f"CHAR{ord(c)}", style=style))

            current_byte_idx += char_width

        return txt

    def _failed_to_decode_msg_txt(self, exception: Optional[Exception]) -> Text:
        """Set `self.failed_to_decode` flag and return a `Text` object with the error message."""
        self.failed_to_decode = True
        return prefix_with_style(f"(decode failed: {exception})", style='red dim italic')



"""
`RegexMatchMetrics` class.
"""
from collections import defaultdict
from dataclasses import dataclass, field

from yaralyzer.bytes_match import BytesMatch
from yaralyzer.decoding.bytes_decoder import BytesDecoder
from yaralyzer.util.logging import log


@dataclass
class RegexMatchMetrics:
    """
    Class to measure what we enounter as we iterate over all matches of a relatively simple byte level regex.

    Things like how much many of our matched bytes were we able to decode easily vs. by force vs. not at all,
    were some encodings have a higher pct of success than others (indicating part of our mystery data might be
    encoded that way?

    Example:
        "Find bytes between quotes" against a relatively large pool of close to random encrypted binary data.

    Attributes:
        match_count (int): Total number of matches found.
        bytes_matched (int): Total number of bytes matched across all matches.
        matches_decoded (int): Number of matches where we were able to decode at least some of the matched bytes.
        easy_decode_count (int): Number of matches where we were able to decode the matched bytes without forcing.
        forced_decode_count (int): Number of matches where we were only able to decode the matched bytes by forcing.
        undecodable_count (int): Number of matches where we were unable to decode any of the matched bytes.
        skipped_matches_lengths (defaultdict): Dictionary mapping lengths of skipped matches to their counts.
        bytes_match_objs (list): List of `BytesMatch` objects for all matches encountered.
        per_encoding_stats (defaultdict): Dictionary mapping encoding names to their respective `RegexMatchMetrics`.
    """
    match_count: int = 0
    bytes_matched: int = 0
    matches_decoded: int = 0
    easy_decode_count: int = 0
    forced_decode_count: int = 0
    undecodable_count: int = 0
    skipped_matches_lengths: dict[int, int] = field(default_factory=lambda: defaultdict(lambda: 0))
    bytes_match_objs: list[BytesMatch] = field(default_factory=list)
    per_encoding_stats: dict[str, 'RegexMatchMetrics'] = field(default_factory=lambda: defaultdict(lambda: RegexMatchMetrics()))

    def num_matches_skipped_for_being_empty(self) -> int:
        """Number of matches skipped for being empty (0 length)."""
        return self.skipped_matches_lengths[0]

    def num_matches_skipped_for_being_too_big(self) -> int:
        """Number of matches skipped for being too big to decode."""
        return sum({k: v for k, v in self.skipped_matches_lengths.items() if k > 0}.values())

    def tally_match(self, decoder: BytesDecoder) -> None:
        """
        Tally statistics from a `BytesDecoder` after it has processed a match.

        Args:
            decoder (BytesDecoder): The `BytesDecoder` that processed a match.
        """
        log.debug(f"Tallying {decoder.bytes_match} ({len(decoder.decodings)} decodings)")
        self.match_count += 1
        self.bytes_matched += decoder.bytes_match.match_length
        self.bytes_match_objs.append(decoder.bytes_match)

        if not decoder.bytes_match.is_decodable():
            self.skipped_matches_lengths[decoder.bytes_match.match_length] += 1

        for decoding_attempt in decoder.decodings:
            log.debug(f"Tallying decoding for {decoding_attempt.encoding}")
            encoding_stats = self.per_encoding_stats[decoding_attempt.encoding]

            if decoding_attempt.failed_to_decode:
                encoding_stats.undecodable_count += 1
            else:
                encoding_stats.match_count += 1
                encoding_stats.matches_decoded += 1

                if decoding_attempt.was_force_decoded:
                    encoding_stats.forced_decode_count += 1
                else:
                    encoding_stats.easy_decode_count += 1

    def __eq__(self, other):
        for k, v in vars(self).items():
            if v != vars(other)[k]:
                return False

        return True

    def __str__(self):
        return f"<matches: {self.match_count}, " + \
               f"bytes: {self.bytes_matched}, " + \
               f"decoded: {self.matches_decoded} " + \
               f"too_big: {self.num_matches_skipped_for_being_too_big()}, " + \
               f"empty: {self.num_matches_skipped_for_being_empty()}>" + \
               f"empty: {self.undecodable_count}>"



"""
Helper methods to work with strings.
"""
import logging
import re
from functools import partial
from typing import Any, Callable, List

ANSI_COLOR_CODE_REGEX = re.compile(r'\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])')
INDENT_DEPTH = 4
INDENT_SPACES = INDENT_DEPTH * ' '
INDENTED_JOINER = ',\n' + INDENT_SPACES
NUMBER_REGEX = re.compile(r"^[\d.]+$")
NON_WORD_CHAR_REGEX = re.compile(r"[^\w]")

is_falsey = lambda s: str(s).lower() in ['', '0', 'false', 'no']
is_truthy = lambda s: s.lower() in ['1', 'true', 'yes']


def escape_yara_pattern(pattern: str) -> str:
    return pattern.replace('/', '\\/')


def hex_to_string(_string: str) -> str:
    r"""String '0D 0A 25 25 45 4F 46 0D 0A' becomes '\r\n%%EOF\r\n'"""
    return bytearray.fromhex(_string.replace(' ', '')).decode()


def indented(s: str, spaces: int = 4, prefix: str = '') -> str:
    indent = ' ' * spaces
    indent += prefix
    return indent + f"\n{indent}".join(s.split('\n'))


def indented_paragraph(s: str, spaces: int = 4, prefix: str = '') -> str:
    return '\n'.join([indented(line) for line in s.split('\n')])


def is_number(s: str) -> bool:
    return bool(NUMBER_REGEX.match(s))


def line_count(_string: str) -> int:
    return len(_string.split("\n"))


def log_level_for(value: str | int) -> int:
    """Accepts log level strings like WARNING, INFO, etc. as well as custom `TRACE` level."""
    if isinstance(value, int):
        return value
    elif re.match(r"\d+", value):
        return int(value)
    elif value in logging.getLevelNamesMapping():
        return logging.getLevelNamesMapping()[value]
    else:
        raise ValueError(f"'{value}' is not a valid log level!")


def props_string(obj: object, keys: list[str] | None = None, joiner: str = ', ') -> str:
    """Generate a string that shows an object's properties, similar to standard repr()."""
    prefix = joiner if '\n' in joiner else ''
    return prefix + joiner.join(props_strings(obj, keys))


def props_string_indented(obj: object, keys: list[str] | None = None) -> str:
    return props_string(obj, keys, INDENTED_JOINER)


def props_strings(obj: object, keys: list[str] | None = None) -> list[str]:
    """Get props of 'obj' in the format ["prop1=5", "prop2='string'"] etc. (for repr(), mostly)."""
    props = []

    for k in (keys or [k for k in vars(obj).keys()]):
        value = getattr(obj, k)
        value = f"'{value}'" if isinstance(value, str) else value
        props.append(f"{k}={value}")

    return props


def str_join(_list: List[Any], separator: str, func: Callable = str) -> str:
    """
    Return a comma separated list of strings. If func is provided the output of calling
    it on each element of the list will be used instead of str()
    """
    func = func or str
    return separator.join([func(item) for item in _list])


def strip_ansi_colors(ansi_str: str) -> str:
    """Remove ANSI color codes from a string."""
    return ANSI_COLOR_CODE_REGEX.sub('', ansi_str).strip()


comma_join = partial(str_join, separator=', ')
newline_join = partial(str_join, separator='\n')



"""
Methods to handle turning various objects into Rich text/table/etc representations.
"""
from typing import List, Optional, Union

from rich import box
from rich.panel import Panel
from rich.style import Style
from rich.text import Text

from yaralyzer.output.theme import BYTES_BRIGHTEST, BYTES_HIGHLIGHT
from yaralyzer.output.console import console
from yaralyzer.util.logging import highlighter, log, log_console

# Color meter realted constants. Make even sized buckets color coded from blue (cold) to green (go)
METER_COLORS = list(reversed([82, 85, 71, 60, 67, 30, 24, 16]))
METER_INTERVAL = (100 / float(len(METER_COLORS))) + 0.1
# Color meter extra style thresholds (these are assuming a scale of 0-100)
UNDERLINE_CONFIDENCE_THRESHOLD = 90
BOLD_CONFIDENCE_THRESHOLD = 60
DIM_COUNTRY_THRESHOLD = 25

# Text object defaults mostly for table entries
NO_DECODING_ERRORS_MSG = Text('No', style='green4 dim')
DECODING_ERRORS_MSG = Text('Yes', style='dark_red dim')


def dim_if(txt: Union[str, Text], is_dim: bool, style: Union[str, None] = None):
    """Apply 'dim' style if 'is_dim'. 'style' overrides for Text and applies for strings."""
    txt = txt.copy() if isinstance(txt, Text) else Text(txt, style=style or '')

    if is_dim:
        txt.stylize('dim')

    return txt


def meter_style(meter_pct: float | int) -> str:
    """For coloring numbers between 0 and 100 (AKA pcts). Closer to 100 means greener, closer to 0.0 means bluer."""
    if meter_pct > 100 or meter_pct < 0:
        log.warning(f"Invalid meter_pct: {meter_pct}")

    color_number = METER_COLORS[int(meter_pct / METER_INTERVAL)]
    style = f"color({color_number})"

    if meter_pct > BOLD_CONFIDENCE_THRESHOLD:
        style += ' bold'
    if meter_pct > UNDERLINE_CONFIDENCE_THRESHOLD:
        style += ' underline'

    return style


def na_txt(style: Union[str, Style] = 'white'):
    """Standard N/A text for tables and such."""
    return Text('N/A', style=style)


def newline_join(texts: List[Text]) -> Text:
    """Join a list of Text objects with newlines between them."""
    return Text("\n").join(texts)


def prefix_with_style(_str: str, style: str, root_style: Optional[Union[Style, str]] = None) -> Text:
    """Sometimes you need a Text() object to start plain lest the underline or whatever last forever."""
    return Text('', style=root_style or 'white') + Text(_str, style)


def print_header_panel(headline: str, style: str, expand: bool = True, padding: tuple | None = None) -> None:
    """
    Print a headline inside a styled Rich `Panel` to the console.

    Args:
        headline (str): The text to display as the panel's headline.
        style (str): The style to apply to the panel (e.g., color, bold, reverse).
        expand (bool, optional): Whether the panel should expand to the full console width. Defaults to `True`.
        padding (tuple, optional): Padding around the panel content (top/bottom, left/right). Defaults to `(0, 2)`.
    """
    console.print(Panel(headline, box=box.DOUBLE_EDGE, expand=expand, padding=padding or (0, 2), style=style))


def reverse_color(style: Style) -> Style:
    """Reverses the color for a given style."""
    return Style(color=style.bgcolor, bgcolor=style.color, underline=style.underline, bold=style.bold)


def size_in_bytes_text(num_bytes: int) -> Text:
    return Text(f"{num_bytes:,d}", 'number').append(' bytes', style='white')


def size_text(num_bytes: int) -> Text:
    """Convert a number of bytes into (e.g.) '54,213 bytes (52 KB)'."""
    kb_txt = prefix_with_style("{:,.1f}".format(num_bytes / 1024), style='bright_cyan', root_style='white')
    kb_txt.append(' kb ')
    bytes_txt = Text('(', 'white') + size_in_bytes_text(num_bytes) + Text(')')
    return kb_txt + bytes_txt


def unprintable_byte_to_text(code: str, style: str = '') -> Text:
    """Used with ASCII escape codes and the like, gives colored results like '[NBSP]'."""
    style = BYTES_HIGHLIGHT if style == BYTES_BRIGHTEST else style
    txt = Text('[', style=style)
    txt.append(code.upper(), style=f"{style} italic dim")
    txt.append(Text(']', style=style))
    return txt



"""
`BytesMatch` class for tracking regex and YARA matches against binary data.
"""
import re
from dataclasses import dataclass, field
from typing import Iterator, Optional

from rich.table import Table
from rich.text import Text
from yara import StringMatch, StringMatchInstance

from yaralyzer.config import YaralyzerConfig
from yaralyzer.util.helpers.rich_helper import prefix_with_style
from yaralyzer.output.file_hashes_table import bytes_hashes_table
from yaralyzer.output.theme import ALERT_STYLE, DEFAULT_HIGHLIGHT_STYLE, GREY_ADDRESS


@dataclass
class BytesMatch:
    """
    Simple class to keep track of regex matches against binary data.

    Basically a Regex `re.match` object with some (not many) extra bells and whistles, most notably
    the `surrounding_bytes` property.

    Args:
        matched_against (bytes): The full byte sequence that was searched.
        start_idx (int): Start index of the match in the byte sequence.
        match_length (int): Length of the match in bytes.
        label (str): Label for the match (e.g., regex or YARA rule name).
        ordinal (int): This was the Nth match for this pattern (used for labeling only).
        match (re.Match, optional): Regex `match` object, if available.
        highlight_style (str, optional): Style to use for highlighting the match.

    Attributes:
        end_idx (int): End index of the match in the byte sequence.
        bytes: (bytes): The bytes that matched the regex.
    """
    matched_against: bytes
    start_idx: int
    match_length: int
    label: str
    ordinal: int
    match: re.Match | None = None   # It's rough to get the regex from yara :(
    highlight_style: str = DEFAULT_HIGHLIGHT_STYLE
    end_idx: int = field(init=False)
    match_grooups: tuple = field(init=False)
    highlight_start_idx: int = field(init=False)
    highlight_end_idx: int = field(init=False)
    surrounding_start_idx: int = field(init=False)
    surrounding_end_idx: int = field(init=False)
    surrounding_bytes: bytes = field(init=False)

    def __post_init__(self):
        self.end_idx: int = self.start_idx + self.match_length
        self.bytes = self.matched_against[self.start_idx:self.end_idx]  # TODO: Maybe should be called "matched_bytes"
        self.match_groups: Optional[tuple] = self.match.groups() if self.match else None
        num_after = YaralyzerConfig.args.surrounding_bytes
        num_before = YaralyzerConfig.args.surrounding_bytes
        # Adjust the highlighting start point in case this match is very early or late in the stream
        self.surrounding_start_idx: int = max(self.start_idx - num_before, 0)
        self.surrounding_end_idx: int = min(self.end_idx + num_after, len(self.matched_against))
        self.surrounding_bytes: bytes = self.matched_against[self.surrounding_start_idx:self.surrounding_end_idx]
        self.highlight_start_idx = self.start_idx - self.surrounding_start_idx
        self.highlight_end_idx = self.highlight_start_idx + self.match_length

    @classmethod
    def from_regex_match(
        cls,
        matched_against: bytes,
        match: re.Match,
        ordinal: int,
        highlight_style: str = DEFAULT_HIGHLIGHT_STYLE
    ) -> 'BytesMatch':
        """
        Alternate constructor to build a `BytesMatch` from a regex match object.

        Args:
            matched_against (bytes): The bytes searched.
            match (re.Match): The regex `match` object.
            ordinal (int): This was the Nth match for this pattern (used for labeling only).
            highlight_style (str, optional): Style for highlighting.

        Returns:
            BytesMatch: The constructed `BytesMatch` instance.
        """
        return cls(matched_against, match.start(), len(match[0]), match.re.pattern, ordinal, match, highlight_style)

    @classmethod
    def from_yara_str(
        cls,
        matched_against: bytes,
        rule_name: str,
        yara_str_match: StringMatch,
        yara_str_match_instance: StringMatchInstance,
        ordinal: int,
        highlight_style: str = DEFAULT_HIGHLIGHT_STYLE
    ) -> 'BytesMatch':
        """
        Alternate constructor to build a `BytesMatch` from a YARA string match instance.

        Args:
            matched_against (bytes): The bytes searched.
            rule_name (str): Name of the YARA rule.
            yara_str_match (StringMatch): YARA string match object.
            yara_str_match_instance (StringMatchInstance): Instance of the string match.
            ordinal (int): The Nth match for this pattern.
            highlight_style (str, optional): Style for highlighting.

        Returns:
            BytesMatch: The constructed BytesMatch instance.
        """
        pattern_label = yara_str_match.identifier

        # Don't duplicate the labeling if rule_name and yara_str are the same
        if pattern_label == '$' + rule_name:
            label = pattern_label
        else:
            label = rule_name + ': ' + pattern_label

        return cls(
            matched_against=matched_against,
            start_idx=yara_str_match_instance.offset,
            match_length=yara_str_match_instance.matched_length,
            label=label,
            ordinal=ordinal,
            highlight_style=highlight_style
        )

    @classmethod
    def from_yara_match(
        cls,
        matched_against: bytes,
        yara_match: dict,
        highlight_style: str = DEFAULT_HIGHLIGHT_STYLE
    ) -> Iterator['BytesMatch']:
        """
        Yield a `BytesMatch` for each string returned as part of a YARA match result dict.

        Args:
            matched_against (bytes): The bytes searched.
            yara_match (dict): YARA match result dictionary.
            highlight_style (str, optional): Style for highlighting.

        Yields:
            BytesMatch: For each string match in the YARA result.
        """
        i = 0  # For numbered labeling

        for yara_str_match in yara_match['strings']:
            for yara_str_match_instance in yara_str_match.instances:
                i += 1

                yield cls.from_yara_str(
                    matched_against,
                    yara_match['rule'],
                    yara_str_match,
                    yara_str_match_instance,
                    i,
                    highlight_style
                )

    def style_at_position(self, idx) -> str:
        """
        Get the style for the byte at position `idx` within the matched bytes.

        Args:
            idx (int): Index within the surrounding bytes.

        Returns:
            str: The style to use for this byte (highlight or greyed out).
        """
        if idx < self.highlight_start_idx or idx >= self.highlight_end_idx:
            return GREY_ADDRESS
        else:
            return self.highlight_style

    def location(self) -> Text:
        """
        Get a styled `Text` object describing the start and end index of the match.

        Returns:
            Text: Rich Text object like '(start idx: 348190, end idx: 348228)'.
        """
        location_txt = prefix_with_style(
            f"(start idx: ",
            style='off_white',
            root_style='decode.subheading'
        )

        location_txt.append(str(self.start_idx), style='number')
        location_txt.append(', end idx: ', style='off_white')
        location_txt.append(str(self.end_idx), style='number')
        location_txt.append(')', style='off_white')
        return location_txt

    def is_decodable(self) -> bool:
        """
        Whether the bytes are decodable depends on whether `SUPPRESS_DECODES_TABLE` is set
        and whether the match length is between `MIN`/`MAX_DECODE_LENGTH`.

        Returns:
            bool: `True` if decodable, `False` otherwise.
        """
        return self.match_length >= YaralyzerConfig.args.min_decode_length \
           and self.match_length <= YaralyzerConfig.args.max_decode_length \
           and not YaralyzerConfig.args.suppress_decodes_table

    def bytes_hashes_table(self) -> Table:
        """
        Build a table of MD5/SHA hashes for the matched bytes.

        Returns:
            Table: Rich `Table` object with hashes.
        """
        return bytes_hashes_table(
            self.bytes,
            self.location().plain,
            'center'
        )

    def suppression_notice(self) -> Text:
        """
        Generate a message for when the match is too short or too long to decode.

        Returns:
            Text: Rich `Text` object with the suppression notice.
        """
        txt = self.__rich__()

        if self.match_length < YaralyzerConfig.args.min_decode_length:
            txt = Text('Too little to actually attempt decode at ', style='grey') + txt
        else:
            txt.append(" too long to decode ")
            txt.append(f"(--max-decode-length is {YaralyzerConfig.args.max_decode_length} bytes)", style='grey')

        return txt

    def to_json(self) -> dict:
        """
        Convert this `BytesMatch` to a JSON-serializable dictionary.

        Returns:
            dict: Dictionary representation of the match, suitable for JSON serialization.
        """
        json_dict = {
            'label': self.label,
            'match_length': self.match_length,
            'matched_bytes': self.bytes.hex(),
            'ordinal': self.ordinal,
            'start_idx': self.start_idx,
            'end_idx': self.end_idx,
            'surrounding_bytes': self.surrounding_bytes.hex(),
            'surrounding_start_idx': self.surrounding_start_idx,
            'surrounding_end_idx': self.surrounding_end_idx,
        }

        if self.match:
            json_dict['pattern'] = self.match.re.pattern

        return json_dict

    def __rich__(self) -> Text:
        """Get a rich `Text` representation of the match for display."""
        headline = prefix_with_style(str(self.match_length), style='number', root_style='decode.subheading')
        headline.append(f" bytes matching ")
        headline.append(f"{self.label} ", style=ALERT_STYLE if self.highlight_style == ALERT_STYLE else 'regex')
        headline.append('at ')
        return headline + self.location()

    def __str__(self):
        """Plain text (no rich colors) representation of the match for display."""
        return self.__rich__().plain



import signal
import functools


def timeout(seconds=5, default=None):
    """
    From: https://stackoverflow.com/questions/75928586/how-to-stop-the-execution-of-a-function-in-python-after-a-certain-time

    Example:
        @timeout(seconds=5, default=None)
            def function():
                pass
    """
    def decorator(func):

        @functools.wraps(func)
        def wrapper(*args, **kwargs):

            def handle_timeout(signum, frame):
                raise TimeoutError()

            signal.signal(signal.SIGALRM, handle_timeout)
            signal.alarm(seconds)
            result = func(*args, **kwargs)
            signal.alarm(0)
            return result

        return wrapper

    return decorator


# import time
# from yaralyzer.util.timeout import timeout

# @timeout(seconds=3, default=None)
# def foo():
#     while True:
#         time.sleep(200)



"""
Functions to export Yaralyzer results to various file formats.
"""
from argparse import Namespace
import json
from pathlib import Path
from subprocess import CalledProcessError
from typing import Callable

from yaralyzer.output.theme import YARALYZER_TERMINAL_THEME
from yaralyzer.util.constants import INKSCAPE, INKSCAPE_URL
from yaralyzer.util.logging import log, log_console, log_file_export
from yaralyzer.util.helpers.env_helper import is_cairosvg_installed
from yaralyzer.util.helpers.shell_helper import ShellResult, get_inkscape_version, safe_args
from yaralyzer.yaralyzer import Yaralyzer

CAIROSVG_WARNING_MSG = f"PNG images rendered by CairoSVG may contain issues, especially with tables. " \
                       f"CairoSVG crashes are also not unheard of. " \
                       f"Consider installing {INKSCAPE.title()} if you plan to export a lot of images.\n" \
                       f"{INKSCAPE_URL}"

# Keys are export function names, values are options we always want to use w/that export function
# Not meant for direct access; instead call invoke_rich_export().
_EXPORT_KWARGS = {
    'save_html': {
        'inline_styles': True,
        'theme': YARALYZER_TERMINAL_THEME,
    },
    'save_svg': {
        'theme': YARALYZER_TERMINAL_THEME,
    },
    'save_text': {
        'styles': True,
    },
}


def export_json(yaralyzer: Yaralyzer, args: Namespace) -> Path:
    """
    Export YARA scan results to JSON.

    Args:
        yaralyzer (Yaralyzer): The `Yaralyzer` object containing the results to export.
        export_basepath (Path | None, Optional): Base path to write output to. Should have no file extension.

    Returns:
        Path: File path data was exported to.
    """
    json_export_path = Path(f"{args._export_basepath}.json")

    with log_file_export(json_export_path):
        matches_data = [m.to_json() for m, _decoder in yaralyzer.match_iterator()]

        with open(json_export_path, 'w') as f:
            json.dump(matches_data, f, indent=4)
            return json_export_path


def invoke_rich_export(export_method: Callable, args: Namespace) -> None:
    """
    Announce the export, perform the export, and announce completion.

    Args:
        export_method (Callable): Usually a `Rich.console.save_whatever()` method.
        args (Namespace, optional): Arguments parsed by ArgumeentParser.

    Returns:
        Path: Path data was exported to.
    """
    method_name = export_method.__name__
    extname = 'txt' if method_name == 'save_text' else method_name.split('_')[-1]
    export_file_path = Path(f"{args._export_basepath}.{extname}")

    if method_name not in _EXPORT_KWARGS:
        raise RuntimeError(f"{method_name} is not a valid Rich.console export method!")

    kwargs = _EXPORT_KWARGS[method_name].copy()
    kwargs.update({'clear': False})
    is_svg_export = 'svg' in method_name

    if is_svg_export:
        kwargs.update({'title': export_file_path.name})

    # Invoke the export method
    with log_file_export(export_file_path):
        log.info(f"Invoking rich.console.{method_name}('{export_file_path}') with kwargs: '{kwargs}'...")
        export_method(export_file_path, **kwargs)

    # PNGs are rendered from SVGs
    if is_svg_export and args.export_png:
        with log_file_export(export_file_path.parent.joinpath(export_file_path.stem + '.png')) as png_path:
            render_png(export_file_path, png_path, args)


def render_png(svg_path: Path, png_path: Path, args: Namespace) -> Path | None:
    """Turn the svg output into a png with Inkscape or cairosvg. Returns png path if successful."""
    try:
        if get_inkscape_version():
            try:
                return _render_png_with_inkscape(svg_path, png_path)
            except (CalledProcessError, FileNotFoundError) as e:
                error_msg = f"Failed to render png with {INKSCAPE}! ({e})"

                if is_cairosvg_installed():
                    log.error(error_msg + f"\n\nFalling back to using cairosvg. Rendered image may me imperfect.")
                else:
                    log.error(error_msg + f"\n\ncairosvg not available to fallback to.")
                    raise e

        return _render_png_with_cairosvg(svg_path, png_path)
    except Exception as e:
        log.error(f"Failed to render png! ({e})")
    finally:
        if not args._keep_exported_svg:
            log.info(f"Removing intermediate SVG file '{svg_path}'...")
            svg_path.unlink()


def _render_png_with_inkscape(svg_path: Path, png_path) -> Path | None:
    log_console.print(f"Rendering .png image with {INKSCAPE}...", highlight=False, style='dim')
    inkscape_cmd = safe_args([INKSCAPE, f'--export-filename={png_path}', svg_path])
    ShellResult.from_cmd(inkscape_cmd, verify_success=True)
    return png_path


def _render_png_with_cairosvg(svg_path: Path, png_path) -> Path:
    log.warning(CAIROSVG_WARNING_MSG)
    import cairosvg
    cairosvg.svg2png(url=str(svg_path), write_to=str(png_path))
    return png_path



"""Main Yaralyzer class and alternate constructors."""
import re
from pathlib import Path
from typing import Callable, Iterator

import yara
from rich.console import Console, ConsoleOptions, RenderResult
from rich.padding import Padding
from rich.style import Style
from rich.text import Text

from yaralyzer.bytes_match import BytesMatch
from yaralyzer.config import YaralyzerConfig
from yaralyzer.decoding.bytes_decoder import BytesDecoder
from yaralyzer.output.file_hashes_table import bytes_hashes_table
from yaralyzer.output.regex_match_metrics import RegexMatchMetrics
from yaralyzer.output.console import console
from yaralyzer.output.theme import DEFAULT_HIGHLIGHT_STYLE, YARALYZER_THEME
from yaralyzer.util.constants import MAX_FILENAME_LENGTH, YARALYZE
from yaralyzer.util.exceptions import print_fatal_error_and_exit
from yaralyzer.util.helpers.file_helper import files_in_dir, load_binary_data, to_paths
from yaralyzer.util.helpers.rich_helper import dim_if, reverse_color
from yaralyzer.util.helpers.string_helper import comma_join, newline_join
from yaralyzer.util.logging import log
from yaralyzer.yara.yara_match import YaraMatch
from yaralyzer.yara.yara_rule_builder import PatternType, YaraModifierType, yara_rule_string

YARA_FILE_DOES_NOT_EXIST_ERROR_MSG = "is not a valid yara rules file (it doesn't exist)"
INVALID_FOR_FILENAME_REGEX = re.compile(r"[^\w,.=+-;]+")


# TODO: might be worth introducing a Scannable namedtuple or similar
class Yaralyzer:
    """
    Central class that handles setting up / compiling YARA rules and reading binary data from files as needed.

    Alternate constructors are provided depending on whether:

    * YARA rules are already compiled

    * YARA rules should be compiled from a string

    * YARA rules should be read from a file

    * YARA rules should be read from a directory of .yara files

    The real action happens in the `__rich__console__()` dunder method.

    Attributes:
        bytes (bytes): The binary data to scan.
        bytes_length (int): The length of the binary data.
        scannable_label (str): A label for the binary data, typically the filename or a user-provided label.
        rules (yara.Rules): The compiled YARA rules to use for scanning.
        rules_label (str): A label for the ruleset, typically derived from filenames or user input.
        highlight_style (str): The style to use for highlighting matches in the output.
        non_matches (list[dict]): A list of YARA rules that did not match the binary data.
        matches (list[YaraMatch]): A list of YaraMatch objects representing the matches found.
        extraction_stats (RegexMatchMetrics): Metrics related to decoding attempts on matched data
    """

    def __init__(
        self,
        rules: str | yara.Rules,
        rules_label: str,
        scannable: bytes | str | Path,
        scannable_label: str | None = None,
        highlight_style: str = DEFAULT_HIGHLIGHT_STYLE
    ) -> None:
        """
        Initialize a `Yaralyzer` instance for scanning binary data with YARA rules.

        Args:
            rules (str | yara.Rules): YARA rules to use for scanning. Can be a string or a pre-compiled
                `yara.Rules` object (strings will be compiled to an instance of `yara.Rules`).
            rules_label (str): Label to identify the ruleset in output and logs.
            scannable (bytes | str | Path): The data to scan. If it's `bytes` type then that data is scanned;
                if it's a string it is treated as a file path to load bytes from.
            scannable_label (str | None, optional): Label for the `scannable` arg data.
                Required if `scannable` is `bytes`. If `scannable` is a Path `scannable_label` will
                default to the file's basename.
            highlight_style (str, optional): Style to use for highlighting matches in output.
                Defaults to `YaralyzerConfig.HIGHLIGHT_STYLE`.

        Raises:
            TypeError: If `scannable` is `bytes` and `scannable_label` is not provided.
        """
        yara.set_config(
            stack_size=YaralyzerConfig.args.yara_stack_size,
            max_match_data=YaralyzerConfig.args.max_match_length
        )

        if isinstance(scannable, (bytes, bytearray, memoryview)):
            if scannable_label is None:
                raise TypeError("Must provide 'scannable_label' arg when yaralyzing raw bytes")

            self.bytes: bytes = scannable
            self.scannable_label: str = scannable_label
        else:
            self.bytes: bytes = load_binary_data(scannable)
            self.scannable_label: str = scannable_label or Path(scannable).name

        if isinstance(rules, yara.Rules):
            self.rules: yara.Rules = rules
        else:
            log.info(f"Compiling YARA rules from provided string:\n{rules}")
            self.rules: yara.Rules = yara.compile(source=rules)

        self.bytes_length: int = len(self.bytes)
        self.rules_label: str = rules_label
        self.highlight_style: str = highlight_style
        # Outcome tracking variables
        self.non_matches: list[dict] = []
        self.matches: list[YaraMatch] = []
        self.extraction_stats = RegexMatchMetrics()

    @classmethod
    def for_rules_files(
        cls,
        yara_rules_files: list[str] | list[Path],
        scannable: bytes | str | Path,
        scannable_label: str | None = None
    ) -> 'Yaralyzer':
        """
        Alternate constructor to load YARA rules from files and label rules with the filenames.

        Args:
            yara_rules_files (list[str]): list of file paths to YARA rules files.
            scannable (Union[bytes, str]): The data to scan. If `bytes`, raw data is scanned;
                if `str`, it is treated as a file path to load bytes from.
            scannable_label (str | None, optional): Label for the `scannable` data.
                Required if `scannable` is `bytes`. If scannable is a file path, defaults to the file's basename.

        Raises:
            FileNotFoundError: If any file in `yara_rules_files` does not exist.
            TypeError: If `yara_rules_files` is not a list of Paths or strings
        """
        yara_rules_paths = to_paths(yara_rules_files)

        for rules_path in yara_rules_paths:
            if not rules_path.exists():
                raise FileNotFoundError(f"'{rules_path}' {YARA_FILE_DOES_NOT_EXIST_ERROR_MSG}")

        try:
            filepaths_arg = {f.name: str(f) for f in yara_rules_paths}
            yara_rules = yara.compile(filepaths=filepaths_arg)
        except yara.SyntaxError as e:
            print_fatal_error_and_exit(f"Failed to parse YARA rules file(s): {e}")

        yara_rules_label = comma_join(sorted([file.name for file in yara_rules_paths]))
        return cls(yara_rules, yara_rules_label, scannable, scannable_label)

    @classmethod
    def for_rules_dirs(
        cls,
        dirs: list[str] | list[Path] | list[str | Path],
        scannable: bytes | str | Path,
        scannable_label: str | None = None
    ) -> 'Yaralyzer':
        """
        Alternate constructor that will load all `.yara` files in `yara_rules_dir`.

        Args:
            dirs (list[str]): list of directories to search for `.yara` files.
            scannable (Union[bytes, str]): The data to scan. If `bytes`, raw data is scanned;
                if `str`, it is treated as a file path to load bytes from.
            scannable_label (str | None, optional): Label for the `scannable` data.
                Required if `scannable` is `bytes`. If scannable is a file path, defaults to the file's basename.

        Raises:
            FileNotFoundError: If `dirs` is not a list of valid directories.
        """
        dirs = to_paths(dirs)

        if not all(dir.is_dir() and dir.exists() for dir in dirs):
            raise FileNotFoundError(f"'{dirs}' is not a list of valid directories")

        rules_files = [f for dir in dirs for f in files_in_dir(dir)]
        log.info(f"Found {len(rules_files)} in {len(dirs)} dirs: {rules_files}")
        return cls.for_rules_files(rules_files, scannable, scannable_label)

    @classmethod
    def for_patterns(
        cls,
        patterns: list[str],
        patterns_type: PatternType,
        scannable: bytes | str | Path,
        scannable_label: str | None = None,
        rules_label: str | None = None,
        pattern_label: str | None = None,
        regex_modifier: YaraModifierType | None = None,
    ) -> 'Yaralyzer':
        """
        Alternate constructor taking regex pattern strings. Rules label defaults to the patterns joined by comma.

        Args:
            patterns (list[str]): list of regex or hex patterns to build rules from.
            patterns_type (PatternType): Either `"regex"` or `"hex"` to indicate the type of patterns provided.
            scannable (Union[bytes, str]): The data to scan. If `bytes`, raw data is scanned;
                if `str`, it is treated as a file path to load bytes from.
            scannable_label (str | None, optional): Label for the `scannable` data.
                Required if `scannable` is `bytes`.
                If scannable is a file path, defaults to the file's basename.
            rules_label (str | None, optional): Label for the ruleset. Defaults to the patterns joined by comma.
            pattern_label (str | None, optional): Label for each pattern in the YARA rules. Defaults to "pattern".
            regex_modifier (str | None, optional): Optional regex modifier (e.g. "nocase", "ascii", "wide", etc).
                Only valid if `patterns_type` is `"regex"`.
        """
        rule_strings = []

        for i, pattern in enumerate(patterns):
            suffix = f"_{i + 1}" if len(patterns) > 1 else ''

            rule_strings.append(yara_rule_string(
                pattern=pattern,
                pattern_type=patterns_type,
                rule_name=f"{rules_label or YARALYZE}{suffix}",
                pattern_label=f"{pattern_label}{suffix}" if pattern_label else None,
                modifier=regex_modifier
            ))

        rules_string = newline_join(rule_strings)
        rules_label = comma_join(patterns)
        return cls(rules_string, rules_label, scannable, scannable_label)

    def export_basepath(self) -> Path:
        """Get the basepath (directory + filename without extension) for exported files."""
        args = YaralyzerConfig.args
        filename_str = INVALID_FOR_FILENAME_REGEX.sub('', str(self).replace(', ', ',').replace(' ', '_'))
        export_basename  = f"{args.file_prefix}{filename_str}"  # noqa: E221
        export_basename += f"__maxdecode{YaralyzerConfig.args.max_decode_length}"
        export_basename += args.file_suffix

        if not args.no_timestamps:
            export_basename += f"__at_{args._invoked_at_str}"

        max_filename_length = MAX_FILENAME_LENGTH - len(str(args.output_dir.resolve()))
        return args.output_dir.joinpath(export_basename[:max_filename_length])

    def yaralyze(self) -> None:
        """Use YARA to find matches and then force decode them."""
        console.print(self)

    def match_iterator(self) -> Iterator[tuple[BytesMatch, BytesDecoder]]:
        """
        Iterator version of `yaralyze()`.

        Yields:
            tuple[BytesMatch, BytesDecoder]: Match and decode data tuple.
        """
        self.rules.match(data=self.bytes, callback=self._yara_callback)

        for yara_match in self.matches:
            console.print(yara_match)
            console.line()

            for match in BytesMatch.from_yara_match(self.bytes, yara_match.match, self.highlight_style):
                decoder = BytesDecoder(match, yara_match.rule_name)
                self.extraction_stats.tally_match(decoder)
                yield match, decoder

        self._print_non_matches()

    def _yara_callback(self, data: dict) -> Callable:
        """
        Callback invoked by `yara-python` to handle matches and non-matches as they are discovered.

        Args:
            data (dict): Data provided when `yara-python` invokes the callback.

        Returns:
            Callable: Always returns `yara.CALLBACK_CONTINUE` to signal `yara-python` should continue processing.
        """
        if data['matches']:
            self.matches.append(YaraMatch(data, self._panel_text()))
        else:
            self.non_matches.append(data)

        return yara.CALLBACK_CONTINUE

    def _print_non_matches(self) -> None:
        """Print info about the YARA rules that didn't match the bytes."""
        if len(self.non_matches) == 0:
            return

        non_matches_text = sorted([Text(nm['rule'], 'grey') for nm in self.non_matches], key=str)

        # Only show the non matches if there were valid ones, otherwise just show the number
        if len(self.matches) == 0:
            non_match_desc = f" did not match any of the {len(self.non_matches)} yara rules"
            console.print(dim_if(self.__text__() + Text(non_match_desc, style='grey'), True))
            return

        non_match_desc = f" did not match the other {len(self.non_matches)} yara rules"
        console.print(self.__text__() + Text(non_match_desc, style='grey') + Text(': '), style='dim')
        console.print(Padding(Text(', ', 'white').join(non_matches_text), (0, 0, 1, 4)))

    def _panel_text(self) -> Text:
        """Inverted colors for the panel at the top of the match section of the output."""
        styles = [reverse_color(YARALYZER_THEME.styles[f"yara.{s}"]) for s in ('scanned', 'rules')]
        return self.__text__(*styles)

    def __text__(self, byte_style: Style | str = 'yara.scanned', rule_style: Style | str = 'yara.rules') -> Text:
        """Text representation of this YARA scan (__text__() was taken)."""
        txt = Text('').append(self.scannable_label, style=byte_style or 'yara.scanned')
        return txt.append(' scanned with <').append(self.rules_label, style=rule_style or 'yara.rules').append('>')

    def __rich_console__(self, _console: Console, options: ConsoleOptions) -> RenderResult:
        """Does the stuff. TODO: not the best place to put the core logic."""
        yield bytes_hashes_table(self.bytes, self.scannable_label)

        for _bytes_match, bytes_decoder in self.match_iterator():
            for attempt in bytes_decoder.__rich_console__(_console, options):
                yield attempt

    def __str__(self) -> str:
        """Plain text (no rich colors) representation of the scan for display."""
        return self.__text__().plain



import re
from argparse import ArgumentParser
from os import getlogin

from yaralyzer.util.helpers.env_helper import stderr_console
from yaralyzer.util.helpers.string_helper import indented, props_string_indented

SKIP_OPTIONS = ['deprecated', 'help', 'option_strings']
STACK_STRIPPER_REGEX = re.compile(fr"/.*?pypoetry/virtualenvs/")


def debug_argparser(parser: ArgumentParser):
    """Debug method to look at argparse internals."""
    from yaralyzer.util.logging import log_console

    for i, action in enumerate((parser)._actions):
        if not action.option_strings:
            continue

        keys = [k for k, v in vars(action).items() if k not in SKIP_OPTIONS and v is not None]
        log_console.print(f"\n{i}: {action.option_strings}", style='cyan', highlight=False)
        log_console.print(props_string_indented(action, keys))


def print_stack():
    import traceback

    for i, stack_obj in enumerate(traceback.extract_stack()):
        line = STACK_STRIPPER_REGEX.sub('', str(stack_obj), 1)
        stderr_console.print(indented(line, i))



"""
Validators for command line arguments.
"""
import re
from abc import ABC
from argparse import ArgumentTypeError
from pathlib import Path

import yara

from yaralyzer.yara.yara_rule_builder import PATTERN_TYPES, build_yara_rule


class OptionValidator(ABC):
    """
    Base class for CLI options validators that needs to be in its own file because of circular
    dependency issues.
    """
    def arg_type_str(self) -> str:
        return type(self).__name__.removesuffix('Validator')


class DirValidator(OptionValidator):
    def __call__(self, value: str) -> Path:
        dir = Path(value)

        if not dir.exists():
            raise ArgumentTypeError(f"'{dir}' is not a directory that exists")
        elif not dir.is_dir():
            raise ArgumentTypeError(f"'{dir}' is a file not directory")

        return dir


class PathValidator(OptionValidator):
    def __call__(self, value: str) -> Path:
        file_path = Path(value)

        if not file_path.exists():
            raise ArgumentTypeError(f"'{file_path}' is not a file that exists")

        return file_path


class PatternsLabelValidator(OptionValidator):
    PATTERN_LABEL_REGEX = re.compile(r"^\w+$")

    def arg_type_str(self) -> str:
        return 'str'

    def __call__(self, value: str) -> str:
        if not self.PATTERN_LABEL_REGEX.match(value):
            raise ArgumentTypeError('--patterns-label can only include alphanumeric and underscore')

        return value


class YaraRegexValidator(OptionValidator):
    def arg_type_str(self) -> str:
        return 'Pattern'

    def __call__(self, value: str) -> str:
        compiled_rule = None
        compilation_error = None

        if '\n' in value:
            raise ArgumentTypeError("Use \\n if you want newlines in your regex")
        elif len(value) > 1 and value[0] == '/' and value[-1] == '/':
            print(f"(hint: you don't need to surround your regex with front slashes)")
            value = value[1:-1]

        # Since we can't know if it's a hex or regex pattern yet, try both and accept if either works.
        for pattern_type in PATTERN_TYPES:
            try:
                compiled_rule = build_yara_rule(value, pattern_type=pattern_type)
            except yara.SyntaxError as e:
                compilation_error = e
            except Exception as e:
                compilation_error = e

        if compiled_rule:
            return value
        else:
            raise ArgumentTypeError(f"SyntaxError: your pattern is invalid ({compilation_error})")



"""
Constants related to character encodings.

Helpful links:

* ISO-8859: [www.mit.edu/people/kenta/two/iso8859.html](https://www.mit.edu/people/kenta/two/iso8859.html)

* UTF-8: [www.utf8-chartable.de/unicode-utf8-table.pl?utf8=dec](https://www.utf8-chartable.de/unicode-utf8-table.pl?utf8=dec)  # noqa: E501
"""

# Bytes (TODO: why is this here?)
NEWLINE_BYTE = b"\n"

# String constants
ENCODING = 'encoding'
ASCII = 'ascii'
UTF_8 = 'utf-8'
UTF_16 = 'utf-16'
UTF_32 = 'utf-32'
ISO_8859_1 = 'iso-8859-1'
WINDOWS_1252 = 'windows-1252'


# Byte order marks
BOMS = {
    b'\x2b\x2f\x76':     'UTF-7 BOM',
    b'\xef\xbb\xbf':     'UTF-8 BOM',
    b'\xfe\xff':         'UTF-16 BOM big-endian',
    b'\xff\xfe':         'UTF-16 BOM little-endian',
    b'\xff\xfe\x00\x00': 'UTF-32 BOM little-endian',
    b'\x00\x00\xfe\xff': 'UTF-32 BOM big-endian',
    b'\x0e\xfe\xff':     'SCSU BOM',
}


# ASCII characters that either print nothing, put the cursor in a weird place, or (worst of all) actively
# delete stuff you already printed
UNPRINTABLE_ASCII = {
    0: 'NUL',    # 'Null',
    1: 'SOH',    # 'StartHeading',
    2: 'STX',    # 'StartText',
    3: 'ETX',
    4: 'EOT',    # End of transmission
    5: 'ENQ',    # 'Enquiry',
    6: 'ACK',    # 'Acknowledgement',
    7: 'BEL',    # 'Bell',
    8: 'BS',     # 'BackSpace',
    # 9:  'HT'   # 'HorizontalTab',
    # 10: 'LF',  # 'LineFeed',
    11: 'VT',    # 'VerticalTab',
    12: 'FF',    # 'FormFeed', AKA 'NewPage'
    13: 'CR',    # 'CarriageReturn',
    14: 'SO',    # 'ShiftOut',
    15: 'SI',    # 'ShiftIn',
    16: 'DLE',   # 'DataLineEscape',
    17: 'DC1',   # DeviceControl1',
    18: 'DC2',   # 'DeviceControl2',
    19: 'DC3',   # 'DeviceControl3',
    20: 'DC4',   # 'DeviceControl4',
    21: 'NAK',   # NegativeAcknowledgement',
    22: 'SYN',   # 'SynchronousIdle',
    23: 'ETB',   # 'EndTransmitBlock',
    24: 'CAN',   # 'Cancel',
    25: 'EM',    # 'EndMedium',
    26: 'SUB',   # 'Substitute',
    27: 'ESC',   # 'Escape',
    28: 'FS',    # 'FileSeparator',
    29: 'GS',    # 'GroupSeparator',
    30: 'RS',    # 'RecordSeparator',
    31: 'US',    # 'UnitSeparator',
    127: 'DEL',  # Delete
}


def scrub_c1_control_chars(char_map: dict) -> None:
    """
    Fill in a `dict` with integer keys/values corresponding to where a given char encoding has no chars
    because this range is for C1 control chars (AKA the "undefined" part of most character maps).
    """
    for i in range(128, 160):
        char_map[i] = f"C1.CHAR{i}"


# ISO-8859-1 AKA "Latin-1". Basically ASCII but using more of 128-256    http://www.gammon.com.au/unicode/
UNPRINTABLE_ISO_8859_1 = UNPRINTABLE_ASCII.copy()
scrub_c1_control_chars(UNPRINTABLE_ISO_8859_1)

UNPRINTABLE_ISO_8859_1.update({
    129: 'HOP',
    141: 'RLF',
    160: 'NBSP',
    173: 'SHY',
})


# UTF-8 Makes no use of 128-256 on their own, only as continuation bytes.
# The C1 bytes can appear but only as continuations
# https://en.wikipedia.org/wiki/UTF-8
UNPRINTABLE_UTF_8 = UNPRINTABLE_ASCII.copy()

# C0, C1, FE, and FF, etc. *never* appear in UTF-8
UNPRINTABLE_UTF_8.update({
    192: 'C0',
    193: 'C1',
    245: 'F5',
    246: 'F6',
    247: 'F7',
    248: 'F8',
    249: 'F9',
    250: 'FA',
    251: 'FB',
    252: 'FC',
    253: 'FD',
    254: 'FE',
    255: 'FF',
})


# Win_1252 is a lot like other 256 char encodings but they colonized the C1 char DMZ in the middle
UNPRINTABLE_WIN_1252 = UNPRINTABLE_ASCII.copy()

UNPRINTABLE_WIN_1252.update({
    129: 'HOP',   # High Octet Preset
    141: 'RLF',   # Reverse Line Feed
    143: 'SS3',   # Single shift 3
    144: 'DCS',   # Device Control String
    147: 'STS',   # Set transmit state
    160: 'NBSP',  # Non-breaking space
})


# ISO-8859-7     #http://www.gammon.com.au/unicode/
UNPRINTABLE_ISO_8859_7 = UNPRINTABLE_ASCII.copy()
scrub_c1_control_chars(UNPRINTABLE_ISO_8859_7)

UNPRINTABLE_ISO_8859_7.update({
    174: 'AE',
    210: 'D2',
    255: 'FF'
})


# Keys are names of encodings we will attempt to decode with, values are dicts mapping the unprintable bytes
# in that encoding to appropriate string represenations of those unprintable bytes.
# Order matters here, as we will attempt the decoding in the order of the keys.
ENCODINGS_TO_ATTEMPT = {
    ASCII:        UNPRINTABLE_ASCII,
    UTF_8:        UNPRINTABLE_UTF_8,
    UTF_16:       None,
    UTF_32:       None,  # UTF-16 and 32 are handled differently
    ISO_8859_1:   UNPRINTABLE_ISO_8859_1,
    WINDOWS_1252: UNPRINTABLE_WIN_1252,
}

SINGLE_BYTE_ENCODINGS = [
    ASCII,
    ISO_8859_1,
    WINDOWS_1252,
]

# Keys are encodings that use multiple bytes to represent a single character, values are the possible offsets
# to attempt to use as the starting point for decoding in a given set of bytes.
WIDE_UTF_ENCODINGS = {
    UTF_16: [0, 1],
    UTF_32: [0, 1, 2, 3],
}


def encoding_offsets(encoding: str) -> list:
    """Get possible offsets for a given encoding. If the encoding is not in `WIDE_UTF_ENCODINGS`, return `[0]`."""
    return WIDE_UTF_ENCODINGS.get(encoding, [0])


def encoding_width(encoding: str) -> int:
    """Get the width of a character in bytes for a given encoding, which is the number of possible offsets."""
    return len(encoding_offsets(encoding))


def is_wide_utf(encoding: str) -> bool:
    """Check if the encoding is a wide UTF encoding (UTF-16 or UTF-32)."""
    return encoding in WIDE_UTF_ENCODINGS


# TODO: this is unused cruft (mostly Asian language encodings)
ENCODINGS = [
    'big5',
    'big5hkscs',
    'cp950',
    'gb2312',
    'gbk',
    'gb18030',
    'hz',
    'iso2022_jp_2',
    'utf-7',
    'utf-8',
    'utf-16',
]



"""
Configuration management for Yaralyzer.
"""
import re
import sys
from contextlib import contextmanager
from copy import deepcopy
from os import environ
from pathlib import Path
from shutil import get_terminal_size
from typing import Any, Generator, Literal, Mapping, Sequence

from dotenv import load_dotenv
from rich.console import Console

from yaralyzer.util.constants import INVOKED_BY_PYTEST, YARALYZER_UPPER, dotfile_name
from yaralyzer.util.helpers.file_helper import relative_path

DEFAULT_CONSOLE_WIDTH = 160
PATH_ENV_VAR_REGEX = re.compile(r"^.*_(DIR|FILE|PATH)S?$", re.I)
PYTEST_REBUILD_FIXTURES_ENV_VAR = 'PYTEST_REBUILD_FIXTURES'
DOTFILE_DIRS = [Path.cwd(), Path.home()]


def config_var_name(env_var: str) -> str:
    """
    Get the name of `env_var` and strip off `YARALYZER_` prefix.

    Example:
        ```
        config_var_name('YARALYZER_SURROUNDING_BYTES') => 'SURROUNDING_BYTES'
        ```
    """
    env_var = env_var.removeprefix(f"{YARALYZER_UPPER}_")
    return f'{env_var=}'.partition('=')[0]


# TODO: why is this a function?
def console_width_possibilities():
    """Returns a list of possible console widths, the first being the current terminal width."""
    # Subtract 2 from terminal cols just as a precaution in case things get weird
    return [get_terminal_size().columns - 2, DEFAULT_CONSOLE_WIDTH]


def is_cairosvg_installed() -> bool:
    """True if cairosvg package is available on the current system."""
    try:
        import cairosvg
        return True
    except ModuleNotFoundError:
        return False


def is_env_var_set_and_not_false(var_name: str) -> bool:
    """Return `True` if `var_name` is not empty and set to anything other than "false" (capitalization agnostic)."""
    if var_name in environ:
        var_value = environ[var_name]
        return var_value is not None and len(var_value) > 0 and var_value.lower() != 'false'
    else:
        return False


def is_invoked_by_pytest() -> bool:
    """Return `True` if invoked in a `pytest` context."""
    return is_env_var_set_and_not_false(INVOKED_BY_PYTEST)


def is_path_var(env_var_name: str) -> bool:
    """Returns True if `env_var_name` ends with _DIR or _PATH."""
    return bool(PATH_ENV_VAR_REGEX.match(env_var_name))


def load_dotenv_file(app_name: Literal['pdfalyzer', 'yaralyzer']) -> None:
    if is_invoked_by_pytest():
        return

    for dotenv_file in [dir.joinpath(dotfile_name(app_name)) for dir in DOTFILE_DIRS]:
        if dotenv_file.exists():
            load_dotenv(dotenv_path=dotenv_file)
            lines = [l for l in dotenv_file.read_text().split('\n') if not l.startswith('#')]
            stderr_console.print(f"Loaded {len(lines)} vars from {relative_path(dotenv_file)}...", style='dim')
            return


@contextmanager
def temporary_argv(new_argv: Sequence[str | Path]) -> Generator[Any, Any, Any]:
    """Temporarily replace sys.argv with something else."""
    old_argv = list(sys.argv)
    sys.argv = [str(arg) for arg in new_argv]

    try:
        yield
    finally:
        sys.argv = old_argv


@contextmanager
def temporary_env(env_vars: Mapping[str, str | Path]) -> Generator[Any, Any, Any]:
    """
    Temporarily add variables to the environemnt.
    See: https://shay-palachy.medium.com/temp-environment-variables-for-pytest-7253230bd777

    Example:
        with temporary_env({'new_var': 'new_value}):
            do_stuff()
    """
    old_environ = dict(environ)
    environ.update({k: str(v) for k, v in env_vars.items()})

    try:
        yield
    finally:
        environ.clear()
        environ.update(old_environ)


def _should_rebuild_fixtures() -> bool:
    """
    True if pytest should overwrite fixture data with new output instead of comparing.
    It's here so that pdfalyzer can also use it.
    """
    return is_env_var_set_and_not_false(PYTEST_REBUILD_FIXTURES_ENV_VAR)


# Maximize output width if YARALYZER_MAXIMIZE_WIDTH is set (also can changed with --maximize-width option)
if is_invoked_by_pytest():
    CONSOLE_WIDTH = DEFAULT_CONSOLE_WIDTH
elif is_env_var_set_and_not_false('YARALYZER_MAXIMIZE_WIDTH'):
    CONSOLE_WIDTH = max(console_width_possibilities())
else:
    CONSOLE_WIDTH = min(console_width_possibilities())

DEFAULT_CONSOLE_KWARGS = {
    'color_system': '256',
    'width': CONSOLE_WIDTH,
}


def default_console_kwargs() -> dict[str, str | int]:
    """Returns a fresh copy of DEFAULT_CONSOLE_KWARGS."""
    return deepcopy(DEFAULT_CONSOLE_KWARGS)


# For use when you need to write output before the main rich.console has managed to get set up.

stderr_console = Console(stderr=True, **{**DEFAULT_CONSOLE_KWARGS, 'width': max(console_width_possibilities())})
# stderr_console.print(f"\n\n MAX WIDTH = {max(console_width_possibilities())}", style='bright_cyan')
